# by Xiaoshuai Jet Zhang <i@jetd.me>, Jan 2021
# v2021.10.08
#
# Depth simulation toolkit.
#
# Please run:
#  1. `pip3 install --upgrade pip`
#  2. `pip3 install opencv-contrib-python scipy open3d`
# before using this package.
#
# TODO:
#  1. IR intensity instead of pixel values in real pipeline.
#

from typing import Optional, Tuple

import numpy as np

# try:
#     import scipy.signal
# except ImportError:
#     print('Please install scipy with `pip3 install scipy`')
#     raise

try:
    import cv2
    from cv2 import ximgproc
except ImportError:
    print("opencv-contrib not installed, " "some features will be disabled.")
    print("Please install with `pip3 install opencv-contrib-python`")
    ximgproc = None


def pad_lr(img: np.ndarray, ndisp: int) -> np.ndarray:
    padding = np.zeros((img.shape[0], ndisp), dtype=np.uint8)
    return np.concatenate([padding, img, padding], axis=1)


def unpad_lr(img: np.ndarray, ndisp: int) -> np.ndarray:
    return img[:, ndisp:-ndisp]


def sim_ir_noise(
    img: np.ndarray,
    scale: float = 0.0,
    blur_ksize: int = 0,
    blur_ksigma: float = 0.03,
    speckle_shape: float = 398.12,
    speckle_scale: float = 2.54e-3,
    gaussian_mu: float = -0.231,
    gaussian_sigma: float = 0.83,
    seed: int = 0,
) -> np.ndarray:
    """
    TODO: IR density model

    Simulate IR camera noise.

    Noise model from Landau et al.
    Simulating Kinect Infrared and Depth Images

    :param img: Input IR image
    :param scale: Scale for downsampling & applying gaussian blur
    :param blur_ksize: Kernel size for gaussian blur
    :param blur_ksigma: Kernel sigma for gaussian blur
    :param speckle_shape: Shape parameter for speckle noise (Gamma distribution)
    :param speckle_scale: Scale parameter for speckle noise (Gamma distribution)
    :param gaussian_mu: mu for additive gaussian noise
    :param gaussian_sigma: sigma for additive gaussian noise
    :param seed: random seed used for numpy
    :return: Processed IR image
    """
    h, w = img.shape

    if scale > 0:
        # downsampling (to emulate soft intensity)
        inter = cv2.INTER_BICUBIC if scale > 1 else cv2.INTER_LANCZOS4
        img = cv2.resize(img, (int(w * scale), int(h * scale)), interpolation=inter)
        if blur_ksize > 0:
            img = cv2.GaussianBlur(img, (blur_ksize, blur_ksize), blur_ksigma)
        img = cv2.resize(img, (w, h), interpolation=cv2.INTER_CUBIC)

    rng = np.random.default_rng(seed)
    img = img.astype(float)

    # speckle noise
    img = img * rng.gamma(shape=speckle_shape, scale=speckle_scale, size=img.shape)

    # gaussian noise
    img = img + gaussian_mu + gaussian_sigma * rng.standard_normal(img.shape)

    # renormalize
    img[img < 0] = 0
    img[img > 255] = 255
    img = img.astype(np.uint8)

    return img


def depth_post_processing(depth: np.ndarray, ksize: int = 5) -> np.ndarray:
    # depth = scipy.signal.medfilt2d(depth, kernel_size=ksize)
    return depth


def get_census(img, wsize=7):
    h, w = img.shape
    assert wsize % 2 == 1

    whalf = wsize // 2
    center = img[whalf : h - whalf, whalf : w - whalf]

    census = np.zeros((h - 2 * whalf, w - 2 * whalf), dtype=np.uint8)
    offsets = [
        (u, v) for v in range(wsize) for u in range(wsize) if not u == v == whalf
    ]

    for u, v in offsets:
        census = (census << 1) | (
            img[v : v + h - 2 * whalf, u : u + w - 2 * whalf] >= center
        )

    ret = np.zeros((h, w), dtype=np.uint8)
    ret[whalf:-whalf, whalf:-whalf] = census

    return ret


def calc_disparity(
    imgl: np.ndarray,
    imgr: np.ndarray,
    method: str,
    *,
    ndisp: int = 96,
    min_disp: int = 0,
    lr_consistency: bool = True,
    use_census: bool = True,
    census_wsize: int = 7,
) -> np.ndarray:
    """
    Calculate disparity given a rectified image pair.

    :param imgl: Left image
    :param imgr: Right image
    :param method: SGBM or BM
    :param ndisp: max disparity
    :param min_disp: min disparity
    :param lr_consistency: Use Left-Right Consistency (LRC) check
    :return: disparity
    """
    if use_census:
        imgl = get_census(imgl, census_wsize)
        imgr = get_census(imgr, census_wsize)

    imgl = pad_lr(imgl, ndisp)
    imgr = pad_lr(imgr, ndisp)

    if method == "SGBM":
        window_size = 7
        matcherl = cv2.StereoSGBM_create(
            minDisparity=min_disp,
            numDisparities=ndisp,
            blockSize=window_size,
            P1=8 * 1 * window_size**2,
            P2=32 * 1 * window_size**2,
            disp12MaxDiff=1,
            uniquenessRatio=10,
            speckleWindowSize=100,
            speckleRange=1,
            mode=cv2.STEREO_SGBM_MODE_HH,
        )
    elif method == "BM":
        matcherl = cv2.StereoBM_create(numDisparities=ndisp, blockSize=7)
    else:
        raise NotImplementedError(f"Not implemented: {method}")

    displ = matcherl.compute(imgl, imgr)

    if lr_consistency:
        if ximgproc is not None:
            matcherr = ximgproc.createRightMatcher(matcherl)
            dispr = matcherr.compute(imgr, imgl)

            wls_filter = ximgproc.createDisparityWLSFilter(matcherl)
            wls_filter.setLRCthresh(16)
            wls_filter.setLambda(0.02)
            wls_filter.setSigmaColor(1.0)
            displ = wls_filter.filter(
                disparity_map_left=displ,
                disparity_map_right=dispr,
                left_view=imgl,
                right_view=imgr,
            )
            # confidence = wls_filter.getConfidenceMap()
        else:
            print("opencv-contrib not installed, post-processing disabled.")

    displ = unpad_lr(displ, ndisp).astype(np.float32) / 16.0

    return displ


def calc_rectified_stereo_pair(
    imgl: np.ndarray,
    imgr: np.ndarray,
    kl: np.ndarray,
    kr: np.ndarray,
    rt: np.ndarray,
    distortl: Optional[np.ndarray] = None,
    distortr: Optional[np.ndarray] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Rectify an image pair with given camera parameters.

    :param imgl: Left image
    :param imgr: Right image
    :param kl: Left intrinsic matrix
    :param kr: Right intrinsic matrix
    :param rt: Extrinsic matrix (left to right)
    :param distortr: Left distortion coefficients
    :param distortl: Right distortion coefficients
    :return imgl_rect: Rectified left image
    :return imgr_rect: Rectified right image
    :return q: Perspective transformation matrix (for cv2.reprojectImageTo3D)
    """
    assert imgl.shape == imgr.shape
    h, w = imgl.shape

    rt = np.linalg.inv(rt)
    # print(rt[:3, :3])
    # print(rt[:3, 3:])
    # print(kl)
    # print(kr)
    # print((w, h))
    # print(distortl)
    # print(distortr)
    r1, r2, p1, p2, q, _, _ = cv2.stereoRectify(
        R=rt[:3, :3],
        T=rt[:3, 3:],
        cameraMatrix1=kl,
        cameraMatrix2=kr,
        alpha=1.0,
        imageSize=(w, h),
        newImageSize=(w, h),
        distCoeffs1=distortl,
        distCoeffs2=distortr,
    )
    # r1, r2, p1, p2, q, _, _ = cv2.stereoRectify(
    #     R=np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]]), T=np.array([[-0.0545], [0.], [0.]]),
    #     cameraMatrix1=np.array([[920., 0., 640.], [0., 920., 360.], [0., 0., 1.]]), cameraMatrix2=np.array([[920., 0., 640.], [0., 920., 360.], [0., 0., 1.]]),
    #     alpha=1.0, imageSize=(w, h), newImageSize=(w, h),
    #     distCoeffs1=distortl, distCoeffs2=distortr
    # )
    map1 = cv2.initUndistortRectifyMap(kl, distortr, r1, p1, (w, h), cv2.CV_32F)
    map2 = cv2.initUndistortRectifyMap(kr, distortl, r2, p2, (w, h), cv2.CV_32F)
    imgl_rect = cv2.remap(imgl, *map1, cv2.INTER_LINEAR)
    imgr_rect = cv2.remap(imgr, *map2, cv2.INTER_LINEAR)

    return imgl_rect, imgr_rect, q


def calc_depth_and_pointcloud(
    disparity: np.ndarray, mask: np.ndarray, q: np.ndarray, no_pointcloud: bool = False
) -> Tuple[np.ndarray, object]:  # object is o3d.geometry.PointCloud
    """
    Calculate depth and pointcloud.

    :param disparity: Disparity
    :param mask: Valid mask
    :param q: Perspective transformation matrix
    :param no_pointcloud:
    :return depth: Depth
    :return pointcloud: Pointcloud
    """
    try:
        import open3d as o3d
    except ImportError:
        print("Please install open3d with `pip3 install open3d`")
        raise

    _3d_image = cv2.reprojectImageTo3D(disparity, q)
    depth = _3d_image[..., 2]
    depth[~mask] = 0
    depth[np.isinf(depth)] = 0
    depth[np.isnan(depth)] = 0

    depth = depth_post_processing(depth) * mask
    _3d_image[..., 2] = depth

    if no_pointcloud:
        pointcloud = None
    else:
        points = _3d_image.reshape(-1, 3)
        valid_flag = mask.reshape(-1)
        valid_points = []
        for i, point in enumerate(points):
            if valid_flag[i]:
                valid_points.append(point)
        valid_points = o3d.utility.Vector3dVector(np.array(valid_points))
        pointcloud = o3d.geometry.PointCloud(points=valid_points)

    return depth, pointcloud


def calc_main_depth_from_left_right_ir(
    ir_l: np.ndarray,
    ir_r: np.ndarray,
    rt_l: np.ndarray,
    rt_r: np.ndarray,
    rt_main: np.ndarray,
    k_l: np.ndarray,
    k_r: np.ndarray,
    k_main: np.ndarray,
    method: str = "SGBM",
    ndisp: int = 96,
    use_noise: bool = True,
    use_census: bool = True,
    lr_consistency: bool = False,
    register_depth: bool = True,
    register_blur_ksize: int = 5,
    main_cam_size=(1920, 1080),
    census_wsize=7,
    **kwargs,
) -> np.ndarray:
    """
    Calculate depth for rgb camera from left right ir images.

    :param ir_l: left ir image
    :param ir_r: right ir image
    :param rt_l: left extrinsic matrix
    :param rt_r: right extrinsic matrix
    :param rt_main: rgb extrinsic matrix
    :param k_l: left intrinsic matrix
    :param k_r: right intrinsic matrix
    :param k_main: rgb intrinsic matrix
    :param method: method for depth calculation (SGBM or BM)
    :param use_noise: whether to simulate ir noise before processing
    :param lr_consistency: whether to use left-right consistency check
    :return depth: calculated depth
    """
    assert ir_l.shape == ir_r.shape

    # assert np.allclose(k_l, k_r)
    # w, h = k_main[:2, 2]
    # w, h = int(w * 2), int(h * 2)
    w, h = main_cam_size
    rt_lr = rt_l @ np.linalg.inv(rt_r)
    rt_mainl = rt_main @ np.linalg.inv(rt_l)
    if not np.allclose(rt_lr[:3, :3], np.eye(3), atol=1e-6):
        raise RuntimeError("extrinsics contain rotation")
    if not (np.sum(rt_lr[1:3, 3] ** 2) < 2e-4):
        raise RuntimeError(f"extrinsics contain translation {rt_lr[:3, 3]}")
    if use_noise:
        ir_l, ir_r = sim_ir_noise(ir_l, **kwargs), sim_ir_noise(ir_r, **kwargs)
    _, _, q = calc_rectified_stereo_pair(
        ir_l, ir_r, k_l.astype(float), k_r.astype(float), rt_lr.astype(float)
    )

    disp = calc_disparity(
        ir_l,
        ir_r,
        method,
        lr_consistency=lr_consistency,
        ndisp=ndisp,
        use_census=use_census,
        census_wsize=census_wsize,
    )
    valid_mask = disp >= 1
    depth, _ = calc_depth_and_pointcloud(disp, valid_mask, q, no_pointcloud=True)

    depth[np.isnan(depth)] = 0
    depth[np.isinf(depth)] = 0
    depth[depth < 0] = 0
    if register_depth:
        depth = cv2.rgbd.registerDepth(
            k_l.astype(float),
            k_main.astype(float),
            None,
            rt_mainl.astype(float),
            depth,
            (w, h),
            depthDilation=True,
        )
        depth[np.isnan(depth)] = 0
        depth[np.isinf(depth)] = 0
        depth[depth < 0] = 0
        if register_blur_ksize > 0:
            depth = cv2.medianBlur(depth, register_blur_ksize)
    return depth


# def get_pointcloud(image_rgb: np.ndarray, image_depth: np.ndarray, intrinsic_matrix: np.ndarray, extrinsic_matrix: np.ndarray):
#     xyz = _depth2pts_np(image_depth.T, intrinsic_matrix)
#     point = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(xyz))
#     colors = image_rgb.transpose([1, 0, 2]).reshape(-1, 3)
#     point.colors = o3d.utility.Vector3dVector(colors.astype(np.float32) / 255.0)
#     point.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
#     point.transform(extrinsic_matrix)
#     return point
#
# def _depth2pts_np(depth_map, cam_intrinsic, cam_extrinsic=np.eye(4)):
#     feature_grid = get_pixel_grids_np(
#         depth_map.shape[0], depth_map.shape[1])
#
#     uv = np.matmul(np.linalg.inv(cam_intrinsic), feature_grid)
#     cam_points = uv * np.reshape(depth_map, (1, -1))
#
#     r = cam_extrinsic[:3, :3]
#     t = cam_extrinsic[:3, 3:4]
#     r_inv = np.linalg.inv(r)
#
#     world_points = np.matmul(r_inv, cam_points - t).transpose()
#     return world_points
#
#
# def get_pixel_grids_np(height, width):
#     x_linspace = np.linspace(0.5, width - 0.5, width)
#     y_linspace = np.linspace(0.5, height - 0.5, height)
#     x_coordinates, y_coordinates = np.meshgrid(x_linspace, y_linspace)
#     x_coordinates = np.reshape(x_coordinates, (1, -1))
#     y_coordinates = np.reshape(y_coordinates, (1, -1))
#     ones = np.ones_like(x_coordinates).astype(float)
#     grid = np.concatenate([x_coordinates, y_coordinates, ones], axis=0)
#     return grid



import pyrfuniverse.attributes as attr


class DigitAttr(attr.BaseAttr):
    """
    Class for simulating DIGIT tactile sensor.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['light']: Bytes of RGB light image in DIGIT.

            self.data['depth']: Bytes of depth image in DIGIT.
        """
        super().parse_message(data)

    def GetData(self):
        """
        Get data from DIGIT in RFUniverse.

        """
        self._send_data("GetData")



# Version of the library that will be used to upload to pypi
__version__ = "0.10.1"

import os.path
import json
from pyrfuniverse.utils.locker import Locker


def read_config():
    if not os.path.exists(config_path):
        print("creating config")
        config = {}
        config["assets_path"] = ""
        config["executable_file"] = ""
        with open(config_path, "w", encoding="utf-8") as file:
            json.dump(config, file, indent=True)
    with open(config_path, "r", encoding="utf-8") as file:
        return json.load(file)


user_path = os.path.expanduser("~/.rfuniverse")
if not os.path.exists(user_path):
    os.makedirs(user_path)
config_path = os.path.join(user_path, "config.json")
with Locker("config"):
    config = read_config()
assets_path = config["assets_path"]
executable_file = config["executable_file"]



import pyrfuniverse.attributes as attr


class HumanbodyAttr(attr.BaseAttr):
    """
    Human body Inverse Kinematic class.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['move_done']: Whether the movement has finished.

            self.data['rotate_done']: Whether the rotation has finished.
        """
        super().parse_message(data)

    def HumanIKTargetDoMove(
        self,
        index: int,
        position: list,
        duration: float,
        speed_based: bool = True,
        relative: bool = False,
    ):
        """
        Human body Inverse Kinematics target movement.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
            position: A list of length 3, representing the position.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `position` is relative; otherwise, `position` is absolute.
        """
        if position is not None:
            assert len(position) == 3, "position length must be 3"
            position = [float(i) for i in position]
        self._send_data(
            "HumanIKTargetDoMove",
            index,
            position,
            float(duration),
            speed_based,
            relative,
        )

    def HumanIKTargetDoRotate(
        self,
        index: int,
        rotation: list,
        duration: float,
        speed_based: bool = True,
        relative: bool = False,
    ):
        """
        Human body Inverse Kinematics target rotation.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
            rotation: A list of length 3, representing the rotation.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `rotation` is relative; otherwise, `rotation` is absolute.
        """
        if rotation is not None:
            assert len(rotation) == 3, "rotation length must be 3"
            rotation = [float(i) for i in rotation]
        self._send_data(
            "HumanIKTargetDoRotate",
            index,
            rotation,
            float(duration),
            speed_based,
            relative,
        )

    def HumanIKTargetDoRotateQuaternion(
        self,
        index: int,
        quaternion: list,
        duration: float,
        speed_based: bool = True,
        relative: bool = False,
    ):
        """
        Human body Inverse Kinematics target rotation using quaternion.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
            quaternion: A list of length 4, representing the quaternion.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `quaternion` is relative; otherwise, `quaternion` is absolute.
        """
        if quaternion is not None:
            assert len(quaternion) == 4, "quaternion length must be 4"
            quaternion = [float(i) for i in quaternion]
        self._send_data(
            "HumanIKTargetDoRotateQuaternion",
            index,
            quaternion,
            float(duration),
            speed_based,
            relative,
        )

    def HumanIKTargetDoComplete(self, index: int):
        """
        Make the human body IK target movement / rotation complete directly.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
        """
        self._send_data("HumanIKTargetDoComplete", index)

    def HumanIKTargetDoKill(self, index: int):
        """
        Make the human body IK target movement / rotation stop.

        Args:
            index: Int, the target for movement. 0 for left hand, 1 for right hand,2 for left foot, 3 for right foot, 4 for head.
        """
        self._send_data("HumanIKTargetDoKill", index)

    def WaitDo(self):
        """
        Wait for the human body IK target movement / rotation complete.
        """
        while not self.data["move_done"] or not self.data["rotate_done"]:
            self.env.step()



from pyrfuniverse.envs import RFUniverseGymGoalWrapper
from pyrfuniverse.utils import RFUniverseToborController
import numpy as np
from gym import spaces
from gym.utils import seeding
import math
import pybullet as p


class ToborPushPullEnv(RFUniverseGymGoalWrapper):
    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        max_steps,
        asset_bundle_file,
        pull=True,
        executable_file=None,
        threshold_angle=None,
    ):
        super().__init__(
            executable_file=executable_file,
            articulation_channel=True,
        )
        self.max_steps = max_steps
        self.pull = pull
        self.asset_bundle_file = asset_bundle_file

        self.object_name_prefix = "Microwave_"
        self.objects = [self.object_name_prefix + str(i + 1) for i in range(15)]
        self.object_position_range_low = np.array([-0.3, 0.4, 1.3])
        self.object_position_range_high = np.array([-0.1, 0.6, 1.5])
        self.tobor_action_space_low = np.array([-0.2, 0, -0.6])
        self.tobor_action_space_high = np.array([0.2, 0.4, -0.3])
        self.eef_orn = np.array([0, 0, math.pi / 2])
        self.episode_object_position = self.object_position_range_high.copy()

        self.threshold_angle = threshold_angle
        if self.threshold_angle is None:
            if self.pull:
                self.threshold_angle = 45.0
            else:
                self.threshold_angle = 15.0

        self.ik_controller = RFUniverseToborController(
            urdf_folder="/home/haoyuan/workspace/tobor",
            left_hand="robotiq85",
            right_hand="robotiq85",
            left_init_joint_positions=[0] * 7,
            right_init_joint_positions=[0] * 7,
        )
        self.seed()
        self.t = 0
        self.goal = self._sample_goal()
        self._reset_object()

        self.action_space = spaces.Box(low=-1, high=1, shape=(5,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )

    def step(self, action: np.ndarray):
        # Position control
        action_ctrl = action.copy()
        pos_ctrl = action_ctrl[:3] * 0.05
        curr_pos = self._get_tobor_eef_position("left")
        pos_ctrl = pos_ctrl + curr_pos
        pos_ctrl = np.clip(
            pos_ctrl,
            self.episode_object_position + self.tobor_action_space_low,
            self.episode_object_position + self.tobor_action_space_high,
        )

        # Rotation control
        rot_ctrl_x = action_ctrl[3] * 5 / math.pi
        curr_rot_x = float(self.eef_orn[0])
        rot_ctrl_x = np.clip(rot_ctrl_x + curr_rot_x, -math.pi / 3, 0)
        self.eef_orn = np.array([rot_ctrl_x, 0, math.pi / 2])

        self._set_tobor_arm(
            mode="left", position=pos_ctrl, eef_euler_angles=self.eef_orn
        )

        # Gripper width control
        gripper_width = self._get_gripper_width("left")
        gripper_width_ctrl = np.clip(gripper_width + action_ctrl[4] * 0.2, 0, 0.085)
        gripper_angle = self._compute_gripper_angle(gripper_width_ctrl)
        self._set_tobor_gripper(mode="left", gripper_angle=gripper_angle)

        self.t += 1
        obs = self._get_obs()
        done = False
        is_success = self._check_success(obs)
        info = {"is_success": is_success}
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)
        if is_success > 0 or self.t == self.max_steps:
            done = True
            obs = self.reset()

        return obs, reward, done, info

    def reset(self):
        super().reset()
        self.env.reset()
        self.ik_controller.reset()

        self.t = 0
        self.goal = self._sample_goal()
        self._destroy_object()
        self._reset_object()

        # Set Tobor arm directly to handle to reduce exploring space
        handle_position = self._get_handle_position()
        if self.pull:
            self.eef_orn = np.array([0, 0, math.pi / 2])
        else:
            self.eef_orn = np.array([-math.pi / 3, 0, math.pi / 2])
        self._set_tobor_arm_directly("left", handle_position, list(self.eef_orn))

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def compute_reward(
        self, achieved_goal: np.ndarray, desired_goal: np.ndarray, info: dict
    ):
        d = self._compute_distance(achieved_goal / 10, desired_goal / 10)
        return -d

    def _get_obs(self):
        tobor_eef_position = self._get_tobor_eef_position("left")
        tobor_eef_orientation = self.eef_orn.copy()
        tobor_eef_width = [self._get_gripper_width("left")]
        tobor_obs = np.concatenate(
            (tobor_eef_position, tobor_eef_orientation, tobor_eef_width)
        )

        handle_position = self._get_handle_position()
        door_axis_position = self._get_door_axis_position()
        open_angle = [self._get_door_open_angle()]
        object_obs = np.concatenate((handle_position, door_axis_position, open_angle))

        obs = np.concatenate((tobor_obs, object_obs))

        return {
            "observation": obs.copy(),
            "achieved_goal": np.array(open_angle),
            "desired_goal": self.goal.copy(),
        }

    def _sample_goal(self):
        return np.array([self.threshold_angle])

    def _destroy_object(self):
        self.articulation_channel.set_action(
            "Destroy",
            index=4,
        )
        self._step()

    def _reset_object(self):
        # object_idx = self.np_random.randint(0, len(self.objects))
        object_idx = 0
        object_position = self.np_random.uniform(
            self.object_position_range_low, self.object_position_range_high
        )
        self.episode_object_position = object_position.copy()
        self.asset_channel.set_action(
            "LoadArticulationBody",
            filename=self.asset_bundle_file,
            name=self.objects[object_idx],
            position=list(object_position),
            rotation=[0, 180, 0],
        )
        self._step()

        # Set init joint position
        if not self.pull:
            self.articulation_channel.set_action(
                "SetJointPositionDirectly", index=4, joint_positions=[60]
            )
        self._step()

    def _get_handle_position(self):
        handle_position = self.articulation_channel.data[4]["positions"][2]
        return np.array(handle_position)

    def _get_door_axis_position(self):
        door_axis_position = self.articulation_channel.data[4]["positions"][1]
        return np.array(door_axis_position)

    def _get_door_open_angle(self):
        return self.articulation_channel.data[4]["joint_positions"][0]

    def _get_tobor_eef_position(self, mode):
        assert mode in ["left", "right"], "Mode is either 'left' or 'right'"
        if mode == "left":
            left_eef_position = self.articulation_channel.data[1]["positions"][11]
            return np.array(left_eef_position)
        else:
            right_eef_position = self.articulation_channel.data[3]["positions"][11]
            return np.array(right_eef_position)

    def _get_gripper_width(self, mode):
        assert mode in ["left", "right"], "Mode is either 'left' or 'right'"
        idx = 1
        if mode == "right":
            idx = 3
        right_inner_finger_pos = np.array(
            self.articulation_channel.data[idx]["positions"][5]
        )
        left_inner_finger_pos = np.array(
            self.articulation_channel.data[idx]["positions"][10]
        )
        width = self._compute_distance(right_inner_finger_pos, left_inner_finger_pos)

        # The position is at the center of inner_finger, so we must get rid of the width of inner finger,
        # to get accurate gripper width
        width = width - 0.00635

        return width

    def _calculate_tobor_arm_joint_positions(self, mode, position, eef_euler_angles):
        assert mode in ["left", "right"], "Mode is either 'left' or 'right'"
        arm_index = 0
        if mode == "left":
            arm_index = 0
        elif mode == "right":
            arm_index = 2

        eef_orn = p.getQuaternionFromEuler(eef_euler_angles)
        joint_positions = self.ik_controller.calculate_ik(mode, position, eef_orn)

        return arm_index, joint_positions

    def _set_tobor_arm(self, mode, position, eef_euler_angles):
        arm_index, joint_positions = self._calculate_tobor_arm_joint_positions(
            mode, position, eef_euler_angles
        )
        self.articulation_channel.set_action(
            "SetJointPosition", index=arm_index, joint_positions=joint_positions
        )
        self._step()

    def _set_tobor_arm_directly(self, mode, position, eef_euler_angles):
        arm_index, joint_positions = self._calculate_tobor_arm_joint_positions(
            mode, position, eef_euler_angles
        )
        self.articulation_channel.set_action(
            "SetJointPositionDirectly", index=arm_index, joint_positions=joint_positions
        )
        self._step()

    def _set_tobor_gripper(self, mode, gripper_angle):
        assert mode in ["left", "right"], "Mode is either 'left' or 'right'"
        gripper_index = 0
        if mode == "left":
            gripper_index = 1
        elif mode == "right":
            gripper_index = 3

        self.articulation_channel.set_action(
            "SetJointPosition",
            index=gripper_index,
            joint_positions=[gripper_angle, gripper_angle],
        )
        self._step()

    def _compute_distance(self, point_a, point_b):
        return np.linalg.norm(point_a - point_b, axis=-1)

    def _compute_gripper_angle(self, width):
        angle_rad = 0.715 - math.asin((width - 0.01) / 0.1143)
        angle_deg = angle_rad * 180 / math.pi

        return angle_deg

    def _check_success(self, obs):
        achieved_goal = obs["achieved_goal"][0]
        desired_goal = obs["desired_goal"][0]

        if self.pull:
            success = (desired_goal < achieved_goal).astype(np.float32)
        else:
            success = (desired_goal > achieved_goal).astype(np.float32)

        return success



from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.side_channel import SideChannel
from pyrfuniverse.side_channel import IncomingMessage
from pyrfuniverse.envs import RFUniverseGymWrapper

import numpy as np
from gym import spaces
from gym.utils import seeding
import uuid


class CollisionAndCleaningDetectionChannel(SideChannel):
    def __init__(self, agents, num_floors=64):
        super().__init__(uuid.UUID("bee25cbc-09e2-11ec-9e67-18c04d443e7d"))
        self._agent_collisions = {}
        self._agent_cleanings = {}
        self._num_collisions = 0
        self._cleaning_states = np.zeros((num_floors,), dtype=bool)
        for agent in agents:
            self._agent_collisions[agent] = 0
            self._agent_cleanings[agent] = 0

    def on_message_received(self, msg: IncomingMessage) -> None:
        self._num_collisions += msg.read_int32()
        self._cleaning_states = np.array([msg.read_float32_list()]) > 0
        for i in range(len(self._agent_collisions.items())):
            agent_name = msg.read_string()
            num_collisions = msg.read_int32()
            num_cleanings = msg.read_int32()

            if agent_name == "":
                continue
            if agent_name[-7:] == "(Clone)":
                agent_name = agent_name[:-7]
            self._agent_collisions[agent_name] += num_collisions
            self._agent_cleanings[agent_name] += num_cleanings

    def get_agent_collisions(self):
        agent_collisions = self._agent_collisions.copy()
        for key, value in self._agent_collisions.items():
            self._agent_collisions[key] = 0

        return agent_collisions

    def get_agent_cleanings(self):
        agent_cleanings = self._agent_cleanings.copy()
        for key, value in self._agent_cleanings.items():
            self._agent_cleanings[key] = 0

        return agent_cleanings

    def get_num_collisions(self):
        num_collisions = self._num_collisions
        self._num_collisions = 0

        return num_collisions

    def get_cleaning_states(self):
        return self._cleaning_states.copy()

    def clear(self):
        self._num_collisions = 0
        for key, value in self._agent_collisions.items():
            self._agent_collisions[key] = 0
            self._agent_cleanings[key] = 0
        self._cleaning_states = np.zeros(self._cleaning_states.shape, dtype=bool)


class CleanerEnv(RFUniverseGymWrapper):
    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        asset_bundle_file,
        max_episode_length=100,
        reset_on_collision=True,
        velocity_reward=False,
        obs_type="multi",
        grid_reward_per_step=False,
        num_agents=3,
        strength=1000,
        collision_multiplier=5,
        num_floors=64,
        executable_file=None,
    ):
        self.num_agents = num_agents
        self.num_floors = num_floors
        self.agent_names = ["agent_" + str(i) for i in range(self.num_agents)]
        self.channel = CollisionAndCleaningDetectionChannel(
            self.agent_names, self.num_floors
        )
        super().__init__(
            executable_file=executable_file,
            custom_channels=[self.channel],
            rigidbody_channel=True,
        )

        self.asset_bundle_file = asset_bundle_file
        self.max_episode_length = max_episode_length
        self.reset_on_collision = reset_on_collision
        self.strength = strength
        self.collision_multiplier = collision_multiplier
        self.velocity_reward = velocity_reward
        self.obs_type = obs_type
        self.grid_reward_per_step = grid_reward_per_step

        self.agent_name = "CleanRobot"
        self.y_offset = 0.1
        self.reset_agent_min_distance = 1.0
        self.world_range_low = np.array([-3, self.y_offset, -3])
        self.world_range_high = np.array([3, self.y_offset, 3])
        self.seed()

        self.action_space = spaces.Box(
            low=-1, high=1, shape=(2 * self.num_agents,), dtype=np.float32
        )

        if self.obs_type == "multi":
            self.observation_space = spaces.Dict(
                {
                    "agent_obs": spaces.Box(
                        low=-np.inf,
                        high=np.inf,
                        shape=(4 * self.num_agents,),
                        dtype=float,
                    ),
                    "floor_obs": spaces.MultiBinary(self.num_floors),
                }
            )
        else:
            self.observation_space = spaces.Box(
                low=-np.inf, high=np.inf, shape=(4 * self.num_agents,), dtype=float
            )
        self._env_setup()
        self.t = 0
        self.num_clean_grids = 0

    def step(self, action: np.ndarray):
        a = action.copy()

        for i in range(self.num_agents):
            target_force = np.array([a[2 * i], 0, a[2 * i + 1]]) * self.strength
            self._set_agent_force(i, target_force)

        done = False
        num_collisions = self.channel.get_num_collisions()
        cleaning_states = self.channel.get_cleaning_states().astype(np.int)
        total_num_clean_grids = np.sum(cleaning_states)

        reward = 0
        if self.grid_reward_per_step:
            assert total_num_clean_grids >= self.num_clean_grids
            reward += total_num_clean_grids - self.num_clean_grids
        else:
            reward += total_num_clean_grids / 10
        if self.velocity_reward:
            reward += self._get_total_velocity() / self.num_agents
        if not self.reset_on_collision:
            reward += -1 * num_collisions * self.collision_multiplier

        info = {"is_success": total_num_clean_grids > self.num_floors * 0.8}
        self.t += 1
        self.num_clean_grids = total_num_clean_grids

        if self.reset_on_collision and num_collisions > 0:
            reward = -1 * self.collision_multiplier * num_collisions
            self.reset()
            done = True

        elif self.t == self.max_episode_length:
            self.reset()
            done = True

        return self._get_obs(), float(reward), done, info

    def reset(self):
        positions = []
        for i in range(self.num_agents):
            while True:
                agent_position = self.np_random.uniform(
                    self.world_range_low, self.world_range_high
                )
                if self._check_reset_position_legality(positions, agent_position):
                    break
            positions.append(agent_position.copy())
            self._set_agent_position(i, agent_position)

        self.t = 0
        self.num_clean_grids = 0
        self.env.reset()
        self.channel.clear()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def _env_setup(self):
        for i in range(self.num_agents):
            agent_position = self.np_random.uniform(
                self.world_range_low, self.world_range_high
            )
            self.asset_channel.set_action(
                "LoadRigidbodyWithName",
                filename=self.asset_bundle_file,
                name=self.agent_name + " " + str(i),
                replace_name=self.agent_names[i],
                position=list(agent_position),
            )
            self._step()

    def _get_obs(self):
        agent_obs = np.array([])
        for i in range(self.num_agents):
            agent_position = self.rigidbody_channel.data[i]["position"]
            agent_velocity = self.rigidbody_channel.data[i]["velocity"]
            # We ignore y-axis
            this_agent_obs = np.array(
                [
                    agent_position[0],
                    agent_position[2],
                    agent_velocity[0],
                    agent_velocity[2],
                ]
            )
            agent_obs = np.concatenate((agent_obs, this_agent_obs))

        floor_obs = self.channel.get_cleaning_states().astype(np.int8)

        if self.obs_type == "multi":
            return {"agent_obs": agent_obs.copy(), "floor_obs": floor_obs.copy()[0]}
        else:
            return agent_obs.copy()

    def _set_agent_position(self, index, position):
        self.rigidbody_channel.set_action(
            "SetTransform", index=index, position=list(position), rotation=[0, 0, 0]
        )
        self._step()

    def _set_agent_force(self, index, force):
        self.rigidbody_channel.set_action("AddForce", index=index, force=force)
        self._step()

    def _get_total_velocity(self):
        total_velocity = 0
        for i in range(self.num_agents):
            velocity = self.rigidbody_channel.data[i]["velocity"]
            total_velocity += np.linalg.norm(
                np.array([velocity[0], velocity[2]]), axis=-1
            )

        return total_velocity

    def _check_reset_position_legality(self, positions, agent_pos):
        for position in positions:
            if np.linalg.norm(position - agent_pos) < self.reset_agent_min_distance:
                return False

        return True



import os


def make_sb3_model_dir(model_dir):
    while str(model_dir).endswith("/"):
        model_dir = model_dir[:-1]
    version = 1
    model_dir_name = model_dir + "_V" + str(version)
    while os.path.exists(model_dir_name):
        version += 1
        model_dir_name = model_dir + "_V" + str(version)

    os.makedirs(model_dir_name)

    return model_dir_name


def get_eval_file_name(model_dir, eval_file_prefix):
    eval_full_path_prefix = os.path.join(model_dir, eval_file_prefix)
    version = 1
    eval_file_name = eval_full_path_prefix + "_V" + str(version) + ".json"
    while os.path.exists(eval_file_name):
        version += 1
        eval_file_name = eval_full_path_prefix + "_V" + str(version) + ".json"

    return eval_file_name



from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.envs import RFUniverseGymGoalWrapper
import numpy as np
from gym import spaces
from gym.utils import seeding
import copy


class NailCardEnv(RFUniverseGymGoalWrapper):
    def __init__(
        self,
        rotation_factor=5,
        vertical_movement_factor=0.01,
        gripper_movement_factor=2,
        nail_movement_factor=0.01,
        goal_baseline=0.05,
        executable_file=None,
    ):
        super().__init__(
            executable_file=executable_file,
            rigidbody_channel=True,
            articulation_channel=True,
            game_object_channel=True,
        )
        self.rotation_factor = rotation_factor
        self.vertical_movement_factor = vertical_movement_factor
        self.gripper_movement_factor = gripper_movement_factor
        self.nail_movement_factor = nail_movement_factor
        self.bit_wise_factor = np.array(
            [
                self.rotation_factor,
                self.vertical_movement_factor,
                self.gripper_movement_factor,
                self.nail_movement_factor,
            ]
        )
        self.goal_baseline = goal_baseline

        self.t = 0
        self.goal = self._sample_goal()
        self.action_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )

    def step(self, a: np.ndarray):
        action = a.copy()
        action_ctrl = action * self.bit_wise_factor
        curr_state = self._get_gripper_extra_param()
        target_state = curr_state + action_ctrl

        self._set_target_state(target_state)
        self.t += 1

        obs = self._get_obs()
        done = False
        info = {"is_success": self._check_success(obs)}
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)

        return obs, reward, done, info

    def reset(self):
        super().reset()
        self.env.reset()
        self._reset_object()

        self.t = 0
        self.goal = self._sample_goal()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        higher_distance = achieved_goal - desired_goal
        return float(-1 * (higher_distance[0] < 0))

    def _get_obs(self):
        gripper_position = np.array(self.articulation_channel.data[0]["positions"][8])
        gripper_velocity = np.array(self.articulation_channel.data[0]["velocities"][8])
        gripper_extra_param = self._get_gripper_extra_param()

        object_pos = np.array(self.rigidbody_channel.data[0]["position"])
        object_rotation = np.array(self.rigidbody_channel.data[0]["rotation"])
        object_velocity = np.array(self.rigidbody_channel.data[0]["velocity"])
        object_angular_vel = np.array(self.rigidbody_channel.data[0]["angular_vel"])

        achieved_goal = np.array([object_pos[1]])

        obs = np.concatenate(
            (
                gripper_position,
                gripper_velocity,
                gripper_extra_param,
                object_pos,
                object_velocity,
                object_rotation,
                object_angular_vel,
            )
        )

        return {
            "observation": obs.copy(),
            "achieved_goal": achieved_goal.copy(),
            "desired_goal": self.goal.copy(),
        }

    def _reset_object(self):
        obj_rot = [0, 90, 0]
        if self.rotation_factor != 0:
            obj_rot[1] = self._generate_random_float(0, 180)
        self.rigidbody_channel.set_action(
            "SetTransform", index=0, position=[0, 0.0025, 0], rotation=obj_rot
        )
        self.game_object_channel.set_action(
            "SetTransform",
            index=0,
            position=[0, self.goal_baseline, 0],
            rotation=obj_rot,
        )
        self._step()

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min < max, "Min value is {}, while max value is {}.".format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float

    def _get_gripper_extra_param(self):
        joint_positions = np.array(self.articulation_channel.data[0]["joint_positions"])
        rotation = joint_positions[0]
        vertical_movement = joint_positions[1]
        finger_width = -1 * (joint_positions[2] + joint_positions[4])
        nail_length = (joint_positions[3] + joint_positions[5]) / 2

        return np.array([rotation, vertical_movement, finger_width, nail_length])

    def _set_target_state(self, target_state):
        finger_joint_position = -1 * target_state[2] / 2
        joint_positions = [
            target_state[0],
            target_state[1],
            finger_joint_position,
            target_state[3],
            finger_joint_position,
            target_state[3],
        ]
        self.articulation_channel.set_action(
            "SetJointPosition", index=0, joint_positions=joint_positions
        )
        self._step()

    def _sample_goal(self):
        return np.array([self.goal_baseline])

    def _check_success(self, obs):
        higher_distance = obs["achieved_goal"] - obs["desired_goal"]
        return (higher_distance > 0).astype(np.float32)



import math
from pyrfuniverse.envs.gym_wrapper_env import RFUniverseGymWrapper
import pyrfuniverse.attributes as attr
from pyrfuniverse.utils.jaco_controller import RFUniverseJacoController
import numpy as np
from gym import spaces
from gym.utils import seeding
import pybullet as p


class KinovaGen2CatchingClothEnv(RFUniverseGymWrapper):
    metadata = {"render.modes": ["human", "rgb_array"]}
    scale = 8
    y_bound = 3.8
    kinova_init_y = 4.8
    object2id = {"kinova": 1001, "target": 1000, "cloth": 89212}

    def __init__(
        self,
        urdf_file,
        lock_eef_height=False,
        with_force_zone=False,
        force_zone_intensity=1,
        force_zone_turbulence=10,
        max_steps=100,
        cloth_init_pos_min=(-1.5, 8, 3.3),
        cloth_init_pos_max=(1.5, 8, 6.3),
        executable_file=None,
        assets=["FallingClothSolver"],
    ):
        super().__init__(
            executable_file=executable_file,
            scene_file="CatchingCloth.json",
            assets=assets,
        )
        self.lock_eef_height = lock_eef_height
        self.with_force_zone = with_force_zone
        self.force_zone_intensity = force_zone_intensity
        self.force_zone_turbulence = force_zone_turbulence
        self.max_steps = max_steps
        self.cloth_init_pos_min = np.array(cloth_init_pos_min)
        self.cloth_init_pos_max = np.array(cloth_init_pos_max)
        self.init_eef_pos = np.array([0, self.kinova_init_y, 4.8]) / self.scale
        self.eef_orn = p.getQuaternionFromEuler(np.array([0, 0, 0]))

        self.seed()
        self._load_cloth()
        self.ik_controller = RFUniverseJacoController(urdf_file)
        self.t = 0
        if not self.lock_eef_height:
            self.action_space = spaces.Box(low=-1, high=1, shape=(3,), dtype=np.float32)
        else:
            self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Box(
            low=-np.inf, high=np.inf, shape=obs.shape, dtype=np.float32
        )

    def step(self, action: np.ndarray):
        delta_pos = action * 0.05
        curr_pos = self._get_eef_position()
        if self.lock_eef_height:
            target_pos = np.array(
                [
                    curr_pos[0] + delta_pos[0],
                    self.kinova_init_y / self.scale,
                    curr_pos[2] + delta_pos[1],
                ]
            )
        else:
            target_pos = curr_pos + delta_pos
        joint_positions = self.ik_controller.calculate_ik_recursive(
            unity_eef_pos=target_pos, eef_orn=self.eef_orn
        )
        self._set_kinova_joints(joint_positions)

        self.t += 1
        obs = self._get_obs()
        reward = self._compute_reward(obs)
        done = False
        fail = self._check_fail(obs)
        if fail or self.t == self.max_steps:
            done = True
        if fail:
            reward -= self.max_steps
        info = {"is_success": done and not fail}

        return obs, reward, done, info

    def reset(self):
        self._destroy_cloth()
        self.ik_controller.reset()
        self.t = 0

        init_joint_positions = self.ik_controller.calculate_ik_recursive(
            unity_eef_pos=self.init_eef_pos, eef_orn=self.eef_orn
        )
        self._set_kinova_joints_directly(init_joint_positions)
        self._load_cloth()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def heuristic(self):
        target_position = self._get_target_position()
        joint_positions = self.ik_controller.calculate_ik_recursive(
            unity_eef_pos=target_position, eef_orn=self.eef_orn
        )
        self._set_kinova_joints_directly(joint_positions)

    def _get_obs(self):
        catcher_position = self._get_eef_position()
        cloth_position = self._get_cloth_position()
        eef_velocity = self._get_eef_velocity()
        cloth_velocity = self._get_cloth_velocity()
        force_zone_parameters = (
            self._get_force_zone_parameters() if self.with_force_zone else np.array([])
        )

        return np.concatenate(
            (
                catcher_position,
                cloth_position,
                eef_velocity,
                cloth_velocity,
                force_zone_parameters,
            )
        )

    def _init_scene(self):
        self.attrs[self.object2id["kinova"]].SetTransform(
            scale=[self.scale, self.scale, self.scale]
        )
        self._step()

    def _load_cloth(self):
        cloth = self.InstanceObject(
            name="FallingClothSolver",
            id=self.object2id["cloth"],
            attr_type=attr.FallingClothAttr,
        )
        cloth.SetTransform(
            position=list(
                self.np_random.uniform(
                    low=self.cloth_init_pos_min, high=self.cloth_init_pos_max
                )
            )
        )
        cloth.SetSolverParameters(gravity=[0, -9.8, 0])
        if self.with_force_zone:
            cloth.SetForceZoneParameters(
                orientation=self.np_random.uniform(-180, 180),
                intensity=self.force_zone_intensity,
                turbulence=self.force_zone_turbulence,
                turbulence_frequency=2,
            )
        self._step()

    def _destroy_cloth(self):
        self.attrs[self.object2id["cloth"]].Destroy()
        self._step()

    def _set_kinova_joints(self, joint_positions):
        joint_positions[7] = 50
        joint_positions[8] = 50
        joint_positions[9] = 50
        self.attrs[self.object2id["kinova"]].SetJointPosition(
            joint_positions=list(joint_positions[0:10])
        )
        self._step()

    def _set_kinova_joints_directly(self, joint_positions):
        joint_positions[7] = 50
        joint_positions[8] = 50
        joint_positions[9] = 50
        self.attrs[self.object2id["kinova"]].SetJointPositionDirectly(
            joint_positions=list(joint_positions[0:10])
        )
        self._step()

    def _get_eef_position(self):
        return (
            np.array(self.attrs[self.object2id["kinova"]].data["positions"][15])
            / self.scale
        )

    def _get_eef_velocity(self):
        return (
            np.array(self.attrs[self.object2id["kinova"]].data["velocities"][15])
            / self.scale
        )

    def _get_target_position(self):
        return (
            np.array(self.attrs[self.object2id["target"]].data["position"]) / self.scale
        )

    def _get_cloth_position(self):
        return (
            np.array(self.attrs[self.object2id["cloth"]].data["avg_position"])
            / self.scale
        )

    def _get_cloth_velocity(self):
        return (
            np.array(self.attrs[self.object2id["cloth"]].data["avg_velocity"])
            / self.scale
        )

    def _get_force_zone_parameters(self):
        return np.array(
            [
                self.attrs[self.object2id["cloth"]].data["force_zone_orientation"]
                / 180
                * math.pi,
                self.attrs[self.object2id["cloth"]].data["force_zone_intensity"],
                self.attrs[self.object2id["cloth"]].data["force_zone_turbulence"],
            ]
        )

    def _check_fail(self, obs):
        cloth_position = obs[3:6]
        return cloth_position[1] < self.y_bound / self.scale

    def _compute_reward(self, obs):
        catcher_position = obs[:3]
        cloth_position = obs[3:6]
        eef_velocity = np.array(obs[6:9])
        cloth_velocity = np.array(obs[9:12])
        """
        Reward constrains the catcher to get closer to cloth as well as maximizing the cloth height and
        minimizing the velocity of end effector. The velocity of end effector should be considered more.
        """

        reward = 0
        # reward += np.tanh(10.0 * cloth_position[1])
        # reward += 1 - np.tanh(10.0 * self._distance(catcher_position, cloth_position))
        reward += 1 - np.tanh(10.0 * abs(cloth_velocity[1]))
        reward += 1 - np.tanh(10.0 * np.linalg.norm(eef_velocity))
        return reward

    def _distance(self, pos1, pos2):
        return np.linalg.norm(pos1 - pos2, axis=-1)



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv

try:
    import gym
except ImportError:
    print("This feature requires gym, please install with `pip install gym==0.21.0`")
    raise


class RFUniverseGymWrapper(RFUniverseBaseEnv, gym.Env):
    def __init__(
        self,
        executable_file: str = None,
        scene_file: str = None,
        assets: list = [],
        **kwargs
    ):
        RFUniverseBaseEnv.__init__(
            self,
            executable_file=executable_file,
            scene_file=scene_file,
            assets=assets,
            **kwargs,
        )

    def close(self):
        RFUniverseBaseEnv.close(self)



import numpy as np
import pyrfuniverse.attributes as attr
import base64


class CameraAttr(attr.BaseAttr):
    """
    Camera attribute class, which can capture many kinds of screenshot
    of the scene in RFUniverse.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['width']: The width of image.

            self.data['height']: The height of image.

            self.data['fov']: The field of view of camera.

            self.data['rgb']: The bytes of rgb image.

            self.data['normal']: The bytes of normal image.

            self.data['id_map']: The bytes of instance segmentation mask image.

            self.data['depth']: The bytes of depth image.

            self.data['depth_exr']: The bytes of depth image in exr format.

            self.data['amodal_mask']: The bytes of amodal mask image.

            self.data['heat_map']: The bytes of heat map image.

            self.data['2d_bounding_box']: The 2d bouding box of objects in camera (image) coordinate.

            self.data['3d_bounding_box']: The 3d bounding box of objects in world coordinate.
        """
        super().parse_message(data)
        if "rgb" in self.data:
            self.data["rgb"] = base64.b64decode(self.data["rgb"])
        if "normal" in self.data:
            self.data["normal"] = base64.b64decode(self.data["normal"])
        if "id_map" in self.data:
            self.data["id_map"] = base64.b64decode(self.data["id_map"])
        if "depth" in self.data:
            self.data["depth"] = base64.b64decode(self.data["depth"])
        if "depth_exr" in self.data:
            self.data["depth_exr"] = base64.b64decode(self.data["depth_exr"])
        if "amodal_mask" in self.data:
            self.data["amodal_mask"] = base64.b64decode(self.data["amodal_mask"])
        if "heat_map" in self.data:
            self.data["heat_map"] = base64.b64decode(self.data["heat_map"])

    def AlignView(self):
        """
        Make the camera in RFUniverse align the current view in GUI.
        """
        self._send_data("AlignView")

    def GetRGB(
        self,
        width: int = 512,
        height: int = 512,
        fov: float = 60.0,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the camera RGB image.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data("GetRGB", intrinsic_matrix, int(width), int(height), float(fov))

    def GetNormal(
        self,
        width: int = 512,
        height: int = 512,
        fov: float = 60.0,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the normal image in world coordinate.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data(
            "GetNormal", intrinsic_matrix, int(width), int(height), float(fov)
        )

    def GetID(
        self,
        width: int = 512,
        height: int = 512,
        fov: float = 60.0,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the instance segmentation mask image. The color for each pixel is computed from object ID, see `pyrfuniverse.utils.rfuniverse_utility.GetColorFromID` for more details.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data("GetID", intrinsic_matrix, int(width), int(height), float(fov))

    def GetDepth(
        self,
        zero_dis: float,
        one_dis: float,
        width: int = 512,
        height: int = 512,
        fov: float = 60.0,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the depth image from camera. Since eacg pixel of depth image returned from this function is 8-bit, user should limit the depth range (`zero_dis` and `one_dis`) for more accurate results.

        Args:
            zero_dis: The minimum distance in calculation.
            one_dis: The maximum distance in calculation.
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data(
            "GetDepth",
            float(zero_dis),
            float(one_dis),
            intrinsic_matrix,
            int(width),
            int(height),
            float(fov),
        )

    def GetDepthEXR(
        self,
        width: int = 512,
        height: int = 512,
        fov: float = 60.0,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the depth image from camera. This function returns EXR format image bytes and each pixel is 32-bit.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data(
            "GetDepthEXR", intrinsic_matrix, int(width), int(height), float(fov)
        )

    def GetAmodalMask(
        self,
        target_id: int,
        width: int = 512,
        height: int = 512,
        fov: float = 60,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the amodal mask image for target object.

        Args:
            target_id: The target object ID.
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data(
            "GetAmodalMask",
            int(target_id),
            intrinsic_matrix,
            int(width),
            int(height),
            float(fov),
        )

    def StartHeatMapRecord(self, targets_id: list):
        targets_id = [int(i) for i in targets_id]

        self._send_data("StartHeatMapRecord", targets_id)

    def EndHeatMapRecord(self):
        self._send_data("EndHeatMapRecord")

    def GetHeatMap(
        self,
        width: int = 512,
        height: int = 512,
        radius: int = 50,
        fov: float = 60.0,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the heat map image.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            radius: The radius of heat map.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data(
            "GetHeatMap",
            int(radius),
            intrinsic_matrix,
            int(width),
            int(height),
            float(fov),
        )

    def Get2DBBox(
        self,
        width: int = 512,
        height: int = 512,
        fov: float = 60.0,
        intrinsic_matrix: np.ndarray = None,
    ):
        """
        Get the 2d bounding box of objects in current camera view.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            radius: The radius of heat map.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data(
            "Get2DBBox", intrinsic_matrix, int(width), int(height), float(fov)
        )

    def Get3DBBox(self):
        """
        Get the 3d bounding box of objects in world coordinate.
        """
        self._send_data("Get3DBBox")



import base64
import cv2
import numpy as np
import pyrfuniverse.attributes as attr
import pyrfuniverse.utils.active_depth_generate as active_depth


class ActiveLightSensorAttr(attr.CameraAttr):
    """
    Class of IR-based depth sensor, which can simulate the noise of
    real-world depth camera and produce depth image in similar pattern.
    """

    def __init__(self, env, id: int, data=None):
        super().__init__(env, id, data)
        self.main_intrinsic_matrix = np.eye(3)
        self.ir_intrinsic_matrix = np.eye(3)

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.\n
            self.data['active_depth']: IR-based depth, shape = (w,h)
        """
        super().parse_message(data)
        if "ir_left" in self.data and "ir_right" in self.data:
            self.data["ir_left"] = base64.b64decode(self.data["ir_left"])
            self.data["ir_right"] = base64.b64decode(self.data["ir_right"])

            image_left = np.frombuffer(self.data["ir_left"], dtype=np.uint8)
            image_left = cv2.imdecode(image_left, cv2.IMREAD_COLOR)[..., 2]
            image_right = np.frombuffer(self.data["ir_right"], dtype=np.uint8)
            image_right = cv2.imdecode(image_right, cv2.IMREAD_COLOR)[..., 2]
            left_extrinsic_matrix = np.array(
                [
                    [0.0, -1.0, 0.0, -0.0175],
                    [0.0, 0.0, -1.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )
            right_extrinsic_matrix = np.array(
                [
                    [0.0, -1.0, 0.0, -0.072],
                    [0.0, 0.0, -1.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )
            main_extrinsic_matrix = np.array(
                [
                    [0.0, -1.0, 0.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
            )
            self.data["active_depth"] = active_depth.calc_main_depth_from_left_right_ir(
                image_left,
                image_right,
                left_extrinsic_matrix,
                right_extrinsic_matrix,
                main_extrinsic_matrix,
                self.ir_intrinsic_matrix,
                self.ir_intrinsic_matrix,
                self.main_intrinsic_matrix,
                lr_consistency=False,
                main_cam_size=(
                    self.main_intrinsic_matrix[0, 2] * 2,
                    self.main_intrinsic_matrix[1, 2] * 2,
                ),
                ndisp=128,
                use_census=True,
                register_depth=True,
                census_wsize=7,
                use_noise=False,
            )
            self.data["active_depth"][self.data["active_depth"] > 8.0] = 0
            self.data["active_depth"][self.data["active_depth"] < 0.1] = 0

    def GetActiveDepth(
        self,
        main_intrinsic_matrix_local: np.ndarray,
        ir_intrinsic_matrix_local: np.ndarray,
    ):
        """
        Get IR-based depth image.

        Args:
            main_intrinsic_matrix_local: np.ndarray The intrinsic matrix of main camera.
            ir_intrinsic_matrix_local: np.ndarray The intrinsic matrix of IR-based camera.
        """
        self.main_intrinsic_matrix = main_intrinsic_matrix_local
        self.ir_intrinsic_matrix = ir_intrinsic_matrix_local
        self._send_data("GetActiveDepth", self.ir_intrinsic_matrix)



from pyrfuniverse.envs.gym_goal_wrapper_env import RFUniverseGymGoalWrapper
from pyrfuniverse.utils.ur5_controller import RFUniverseUR5Controller
import numpy as np
import pybullet as p
from gym import spaces
from gym.utils import seeding
import math


class UR5WaterShootingEnv(RFUniverseGymGoalWrapper):
    metadata = {"render.modes": ["human", "rgb_array"]}
    object2id = {
        "ur5": 1001,
        "robotiq85": 10010,
        "target": 1003,
        "right_finger": 100100,
        "left_finger": 100101,
        "cube": 1002,
        "goal": 1000,
    }

    def __init__(
        self,
        urdf_file,
        max_steps=100,
        object_range_low=(-20, 1, -5),
        object_range_high=(-16, 1, 5),
        goal_range_x=(-26, -23),
        goal_range_z=(-3, 3),
        tolerance=2,
        volume_per_time_step_range=(0.05, 0.15),
        liquid_init_velocity_range=(2, 3),
        executable_file=None,
        assets: list = [],
    ):
        super().__init__(
            executable_file=executable_file,
            scene_file="WaterShooting.json",
            assets=assets,
        )

        self.scale = 15
        self.max_steps = max_steps
        self.object_range_low = np.array(object_range_low)
        self.object_range_high = np.array(object_range_high)
        self.goal_range_x = np.array(goal_range_x)
        self.goal_range_z = np.array(goal_range_z)
        self.tolerance = tolerance
        self.volume_per_time_step_range = np.array(volume_per_time_step_range)
        self.liquid_init_velocity_range = np.array(liquid_init_velocity_range)

        self.ik_controller = RFUniverseUR5Controller(
            urdf_file, init_joint_positions=[0, -2 * math.pi / 3, math.pi / 2, 0, 0, 0]
        )
        self.eef_orn = p.getQuaternionFromEuler(
            np.array([-math.pi / 2, 0, -math.pi / 2])
        )
        self.init_pos = np.array([-8, 4, 0]) / self.scale
        self.init_gripper_width = 0.04

        self._env_setup()
        self.seed()
        self.goal = self._sample_goal()

        self.action_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )

    def step(self, action: np.ndarray):
        delta_pos = action[:3].copy() * 0.05
        curr_pos = self._get_eef_position() / self.scale
        target_pos = curr_pos + delta_pos
        joint_positions = self.ik_controller.calculate_ik_recursive(
            unity_eef_pos=target_pos, eef_orn=self.eef_orn
        )

        delta_width = action[3] * 0.02
        curr_width = self._get_gripper_width() / self.scale
        target_width = np.clip(float(curr_width + delta_width), 0.04, 0.085)
        joint_positions.append(target_width)

        self._set_ur5_robotiq85(joint_positions)
        self._set_liquid_parameters(self._get_gripper_width() / self.scale)
        self.t += 1
        obs = self._get_obs()
        info = {
            "is_success": self._check_success(obs["achieved_goal"], obs["desired_goal"])
        }
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)
        done = False
        fail = self._check_fail(obs["achieved_goal"])
        if fail or self.t == self.max_steps or info["is_success"] > 0:
            done = True

        return obs, reward, done, info

    def reset(self):
        self.ik_controller.reset()
        init_joint_positions = self.ik_controller.calculate_ik_recursive(
            unity_eef_pos=self.init_pos, eef_orn=self.eef_orn
        )
        init_joint_positions.append(self.init_gripper_width)
        self._set_ur5_robotiq85_directly(init_joint_positions)

        self.t = 0
        self._reset_liquid()
        object_pos = self._reset_object()
        self.goal = self._sample_goal(object_pos)

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        distance = self._distance(achieved_goal, desired_goal) / self.scale
        return -distance.astype(np.float32)

    def heuristic(self):
        self.ik_controller.reset()
        target_pos = self._get_target_pos() / self.scale
        joint_positions = self.ik_controller.calculate_ik_recursive(
            unity_eef_pos=target_pos, eef_orn=self.eef_orn
        )
        joint_positions.append(0)
        self._set_ur5_robotiq85_directly(joint_positions)

    def _get_obs(self):
        eef_pos = self._get_eef_position()
        eef_vel = self._get_eef_velocity()
        eef_width = self._get_gripper_width()
        robot_obs = np.concatenate((eef_pos, eef_vel, [eef_width]))

        object_pos = self._get_cube_pos()
        object_vel = self._get_cube_velocity()
        object_rot = self._get_cube_rotation()
        object_obs = np.concatenate((object_pos, object_rot, object_vel))

        achieved_goal = object_pos.copy()
        desired_goal = self.goal.copy()

        return {
            "observation": np.concatenate((robot_obs, object_obs)),
            "achieved_goal": achieved_goal,
            "desired_goal": desired_goal,
        }

    def _env_setup(self):
        self._step()

    def _sample_goal(self, cube_pos=None):
        goal_x = self.np_random.uniform(
            low=self.goal_range_x[0], high=self.goal_range_x[1]
        )
        if cube_pos is not None:
            goal_z = cube_pos[2] + self.np_random.uniform(
                low=self.goal_range_z[0], high=self.goal_range_z[1]
            )
        else:
            goal_z = 0

        goal = [float(goal_x), 0, float(goal_z)]
        self.attrs[self.object2id["goal"]].SetTransform(position=list(goal))
        self._step()

        return np.array(goal)

    def _reset_liquid(self):
        self.SendMessage("SetZibraLiquid", False)
        self._step()
        self.SendMessage("SetZibraLiquid", True)
        self._step()

    def _reset_object(self):
        object_pos = self.np_random.uniform(
            low=self.object_range_low, high=self.object_range_high, size=(3,)
        )
        self.attrs[self.object2id["cube"]].SetTransform(
            position=list(object_pos), rotation=[0, 0, 0]
        )
        self._step()
        return object_pos.copy()

    def _get_eef_position(self):
        return np.array(self.attrs[self.object2id["robotiq85"]].data["positions"][7])

    def _get_eef_velocity(self):
        return np.array(self.attrs[self.object2id["robotiq85"]].data["velocities"][7])

    def _get_gripper_width(self):
        left_finger_pos = np.array(
            self.attrs[self.object2id["left_finger"]].data["position"]
        )
        right_finger_pos = np.array(
            self.attrs[self.object2id["right_finger"]].data["position"]
        )
        return self._distance(left_finger_pos, right_finger_pos)

    def _get_target_pos(self):
        return np.array(self.attrs[self.object2id["target"]].data["position"])

    def _get_cube_pos(self):
        return np.array(self.attrs[self.object2id["cube"]].data["position"])

    def _get_cube_velocity(self):
        return np.array(self.attrs[self.object2id["cube"]].data["velocity"])

    def _get_cube_rotation(self):
        return (
            np.array(self.attrs[self.object2id["cube"]].data["rotation"])
            / 180
            * math.pi
        )

    def _set_ur5_robotiq85(self, joint_positions):
        self.attrs[self.object2id["ur5"]].SetJointPosition(
            joint_positions=list(joint_positions[:6])
        )
        self._step()

        width = joint_positions[6]
        gripper_angle = self._compute_gripper_angle(width)
        self.attrs[self.object2id["robotiq85"]].SetJointPosition(
            joint_positions=[gripper_angle, gripper_angle]
        )
        self._step()

    def _set_ur5_robotiq85_directly(self, joint_positions):
        self.attrs[self.object2id["ur5"]].SetJointPositionDirectly(
            joint_positions=list(joint_positions[:6])
        )
        self._step()

        width = joint_positions[6]
        gripper_angle = self._compute_gripper_angle(width)
        self.attrs[self.object2id["robotiq85"]].SetJointPositionDirectly(
            joint_positions=[gripper_angle, gripper_angle]
        )
        self._step()

    def _distance(self, pos1, pos2):
        assert pos2.shape == pos1.shape
        return np.linalg.norm(pos1 - pos2, axis=-1)

    def _compute_gripper_angle(self, width):
        angle_rad = 0.715 - math.asin((width - 0.01) / 0.1143)
        angle_deg = angle_rad * 180 / math.pi

        return angle_deg

    def _set_liquid_parameters(self, gripper_width):
        prop = gripper_width / 0.085
        volume_per_time_step = (
            prop * self.volume_per_time_step_range[1]
            + (1 - prop) * self.volume_per_time_step_range[0]
        )
        liquid_init_velocity = (
            prop * self.liquid_init_velocity_range[1]
            + (1 - prop) * self.liquid_init_velocity_range[0]
        )
        self.SendMessage(
            "SetZibraLiquidEmitter",
            volume_per_time_step,
            0.0,
            -liquid_init_velocity,
            0.0,
        )
        self._step()

    def _check_success(self, achieved_goal, desired_goal):
        distance = abs(achieved_goal - desired_goal)
        return (max(distance[0], distance[2]) < self.tolerance).astype(np.float32)

    def _check_fail(self, achieved_goal):
        if achieved_goal[0] < -28 or abs(achieved_goal[2]) > 10:
            return True
        return False



import numpy as np
import pyrfuniverse.attributes as attr


class ControllerAttr(attr.ColliderAttr):
    """
    Robot controller class, which will control robot arms, hands and embodied robots.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['number_of_joints']: The number of joints in an articulation.

            self.data['positions']: The position of each part in an articulation.

            self.data['rotations']: The rotation of each part in an articulation.

            self.data['quaternion']: The quaternion of each part in an articulation.

            self.data['local_positions']: The local position of each part in an articulation.

            self.data['local_rotations']: The local rotation of each part in an articulation.

            self.data['local_quaternion']: The local quaternion of each part in an articulation.

            self.data['velocities']: The velocity of each part in an articulation.

            self.data['number_of_moveable_joints']: The number of moveable joints in an articulation.

            self.data['joint_positions']: The joint position of each moveable joint in an articulation.

            self.data['joint_velocities']: The joint velocity of each moveable joint in an articulation.

            self.data['all_stable']: Whether all joints have finished moving.

            self.data['move_done']: Whether robot arm IK has finished moving.

            self.data['rotate_done']: Whether robot arm IK has finished rotating.

            self.data['gravity_forces']: Inverse Dynamics force needed to counteract gravity.

            self.data['coriolis_centrifugal_forces']: Inverse Dynamics force needed to counteract coriolis centrifugal forces.

            self.data['drive_forces']: Inverse Dynamics drive forces.
        """
        super().parse_message(data)

    def SetJointPosition(self, joint_positions: list, speed_scales: list = None):
        """
        Set the target joint position for each moveable joint and move with PD control.

        Args:
            joint_positions: A list of float, representing the target joint positions.
            speed_scales: A list of float, representing the speed scale.
        """
        if speed_scales is not None:
            assert len(joint_positions) == len(
                speed_scales
            ), "The length of joint_positions and speed_scales are not equal."
            speed_scales = [float(i) for i in speed_scales]
        joint_positions = [float(i) for i in joint_positions]
        self._send_data("SetJointPosition", joint_positions, speed_scales)

    def SetJointPositionDirectly(self, joint_positions: list):
        """
        Set the target joint position for each moveable joint and move directly.

        Args:
            joint_positions: A list of float, representing the target joint positions.
        """
        joint_positions = [float(i) for i in joint_positions]
        self._send_data("SetJointPositionDirectly", joint_positions)

    def SetIndexJointPosition(self, index: int, joint_position: float):
        """
        Set the target joint position for a given joint and move with PD control.

        Args:
            index: Int, joint index.
            joint_position: Float, the target joint position.
        """
        self._send_data("SetIndexJointPosition", index, float(joint_position))

    def SetIndexJointPositionDirectly(self, index: int, joint_position: float):
        """
        Set the target joint position for a given joint and move directly.

        Args:
            index: Int, joint index.
            joint_position: Float, the target joint position.
        """
        self._send_data("SetIndexJointPositionDirectly", index, float(joint_position))

    def SetJointPositionContinue(self, interval: int, time_joint_positions: list):
        """
        Set the target joint position for each moveable joint and move with PD control continuously.

        Args:
            interval: Float, the time interval.
            time_joint_positions: A list of float list, representing the target joint positions at each time step.
        """
        for i in range(len(time_joint_positions)):
            time_joint_positions[i] = [float(j) for j in time_joint_positions[i]]
        self._send_data("SetJointPositionContinue", interval, time_joint_positions)

    def SetJointVelocity(self, joint_velocitys: list):
        """
        Set the target joint velocity for each moveable joint.

        Args:
            joint_velocitys: A list of float, representing the target joint velocities.
        """
        joint_velocitys = [float(i) for i in joint_velocitys]
        self._send_data("SetJointVelocity", joint_velocitys)

    def SetIndexJointVelocity(self, index: int, joint_velocity: float):
        """
        Set the target joint velocity for a given joint.

        Args:
            index: Int, joint index.
            joint_velocity: A list of float, representing the target joint velocities.
        """
        self._send_data("SetIndexJointVelocity", index, float(joint_velocitys))

    def AddJointForce(self, joint_forces: list):
        """
        Add force to each moveable joint.

        Args:
            joint_forces: A list of forces, representing the added forces.
        """
        joint_forces = [float(i) for i in joint_forces]
        self._send_data("AddJointForce", joint_forces)

    def AddJointForceAtPosition(self, joint_forces: list, force_positions: list):
        """
        Add force to each moveable joint at a given position.

        Args:
            joint_forces: A list of forces, representing the added forces.
            force_positions: A list of positions, representing the positions for forces.
        """
        assert len(joint_forces) == len(
            force_positions
        ), "The length of joint_forces and force_positions are not equal."
        joint_forces = [float(i) for i in joint_forces]
        force_positions = [float(i) for i in force_positions]
        self._send_data("AddJointForceAtPosition", joint_forces, force_positions)

    def AddJointTorque(self, joint_torques: list):
        """
        Add torque to each moveable joint.

        Args:
            joint_torques: A list of torques, representing the added torques.
        """
        joint_torques = [float(i) for i in joint_torques]
        self._send_data("AddJointTorque", joint_torques)

    # only work on unity 2022.1+
    def GetJointInverseDynamicsForce(self):
        """
        Get the joint inverse dynamic force of each moveable joint. Note that this function only works in Unity version >= 2022.1.
        """
        self._send_data("GetJointInverseDynamicsForce")

    def SetImmovable(self, immovable: bool):
        """
        Set whether the base of articulation is immovable.

        Args:
            immovable: Bool, True for immovable, False for movable.
        """
        self._send_data("SetImmovable", immovable)

    def MoveForward(self, distance: float, speed: float):
        """
        Move robot forward. Only works if the robot controller has implemented functions inherited from `ICustomMove.cs`. See https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ICustomMove.cs and https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ToborMove.cs for more details.

        Args:
            distance: Float, distance.
            speed: Float, velocity.
        """
        self._send_data("MoveForward", float(distance), float(speed))

    def MoveBack(self, distance: float, speed: float):
        """
        Move robot backword. Only works if the robot controller has implemented functions inherited from `ICustomMove.cs`. See https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ICustomMove.cs and https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ToborMove.cs for more details.

        Args:
            distance: Float, distance.
            speed: Float, velocity.
        """
        self._send_data("MoveBack", float(distance), float(speed))

    def TurnLeft(self, angle: float, speed: float):
        """
        Turn robot left. Only works if the robot controller has implemented functions inherited from `ICustomMove.cs`. See https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ICustomMove.cs and https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ToborMove.cs for more details.

        Args:
            angle: Float, rotation angle.
            speed: Float, velocity.
        """
        self._send_data("TurnLeft", float(angle), float(speed))

    def TurnRight(self, angle: float, speed: float):
        """
        Turn robot right. Only works if the robot controller has implemented functions inherited from `ICustomMove.cs`. See https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ICustomMove.cs and https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ToborMove.cs for more details.

        Args:
            angle: Float, rotation angle.
            speed: Float, velocity.
        """
        self._send_data("TurnRight", float(angle), float(speed))

    def GripperOpen(self):
        """
        Open the gripper. Only works if the robot controller has implemented functions inherited from `ICustomGripper.cs`. See https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ICustomGripper.cs and https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/GeneralGripper.cs for more details.
        """
        self._send_data("GripperOpen")

    def GripperClose(self):
        """
        Close the gripper. Only works if the robot controller has implemented functions inherited from `ICustomGripper.cs`. See https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/ICustomGripper.cs and https://github.com/mvig-robotflow/rfuniverse/blob/main/Assets/RFUniverse/Scripts/Utils/GeneralGripper.cs for more details.
        """
        self._send_data("GripperClose")

    def EnabledNativeIK(self, enabled: bool):
        """
        Enable or disable the native IK algorithm.

        Args:
            enabled: Bool, True for enable and False for disable.When it is True, through the IKTatGetDo*** interface, according to the end pose.When it is False, through the SetJoint*** interface, according to the joint movement.NativeIK can only take effect when it is started during initialization.
        """
        self._send_data("EnabledNativeIK", enabled)

    def IKTargetDoMove(
        self,
        position: list,
        duration: float,
        speed_based: bool = True,
        relative: bool = False,
    ):
        """
        Native IK target movement.

        Args:
            position: A list of length 3, representing the position.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `position` is relative; otherwise, `position` is absolute.
        """
        if position is not None:
            assert len(position) == 3, "position length must be 3"
            position = [float(i) for i in position]

        self._send_data(
            "IKTargetDoMove", position, float(duration), speed_based, relative
        )

    def IKTargetDoRotate(
        self,
        rotation: list,
        duration: float,
        speed_based: bool = True,
        relative: bool = False,
    ):
        """
        Native IK target rotation.

        Args:
            rotation: A list of length 3, representing the rotation.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `rotation` is relative; otherwise, `rotation` is absolute.
        """
        if rotation is not None:
            assert len(rotation) == 3, "rotation length must be 3"
            rotation = [float(i) for i in rotation]

        self._send_data(
            "IKTargetDoRotate", rotation, float(duration), speed_based, relative
        )

    def IKTargetDoRotateQuaternion(
        self,
        quaternion: list,
        duration: float,
        speed_based: bool = True,
        relative: bool = False,
    ):
        """
        Native IK target rotation using quaternion.

        Args:
            quaternion: A list of length 4, representing the quaternion.
            duration: Float, if `speed_based` is True, it represents movement duration; otherwise, it represents movement speed.
            speed_based: Bool.
            relative: Bool, if True, `quaternion` is relative; otherwise, `quaternion` is absolute.
        """
        if quaternion is not None:
            assert len(quaternion) == 4, "quaternion length must be 4"
            quaternion = [float(i) for i in quaternion]

        self._send_data(
            "IKTargetDoRotateQuaternion",
            quaternion,
            float(duration),
            speed_based,
            relative,
        )

    def IKTargetDoComplete(self):
        """
        Make native IK target movement / rotation complete directly.
        """
        self._send_data("IKTargetDoComplete")

    def IKTargetDoKill(self):
        """
        Make native IK target movement / rotation stop.
        """
        self._send_data("IKTargetDoKill")

    def SetIKTargetOffset(
        self,
        position: list = [0.0, 0.0, 0.0],
        rotation: list = [0.0, 0.0, 0.0],
        quaternion: list = None,
    ):
        """
        Set the new IK target by setting offset to the original target of native IK.

        Args:
            position: A list of length 3, representing the position offset to original target.
            rotation: A list of length 3, representing the rotation offset to original target.
            quaternion: A list of length 4, representing the quaternion offset to original target. If this parameter is specified, `rotation` will be ignored.
        """
        if position is not None:
            assert len(position) == 3, "position length must be 3"
            position = [float(i) for i in position]
        if rotation is not None:
            assert len(rotation) == 3, "rotation length must be 3"
            rotation = [float(i) for i in rotation]
        if quaternion is not None:
            assert len(quaternion) == 4, "quaternion length must be 4"
            quaternion = [float(i) for i in quaternion]

        self._send_data("SetIKTargetOffset", position, rotation, quaternion)

    def WaitDo(self):
        """
        Wait for the native IK target movement / rotation complete.
        """
        while not self.data["move_done"] or not self.data["rotate_done"]:
            self.env._step()



import pyrfuniverse.attributes as attr


class RigidbodyAttr(attr.ColliderAttr):
    """
    Rigid body class.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            self.data['velocity']: The velocity of the object.

            self.data['angular_vel']: The angular velcity of the object.
        """
        super().parse_message(data)

    def SetMass(self, mass: float):
        """
        Set the mass of this rigid body object

        Args:
            mass: Float, representing the mass of this rigid body.
        """
        self._send_data("SetMass", float(mass))

    def AddForce(self, force: list):
        """
        Add force to this rigid body object.

        Args:
            force: A list of length 3, representing the force added to this rigid body.
        """
        if force is not None:
            force = [float(i) for i in force]

        self._send_data("AddForce", force)

    def SetVelocity(self, velocity: list):
        """
        Set the velocity of this rigid body object.

        Args:
            velocity: A list of length 3, representing the velocity of this rigid body.
        """
        if velocity is not None:
            velocity = [float(i) for i in velocity]

        self._send_data("SetVelocity", velocity)

    def SetKinematic(self, is_kinematic: bool):
        """
        Set the Rigidbody is kinematic or not.

        Args:
            is_kinematic: is kinematic or not.
        """
        self._send_data("SetKinematic", is_kinematic)



import pyrfuniverse.attributes as attr


class FallingClothAttr(attr.BaseAttr):
    def parse_message(self, data: dict):
        super().parse_message(data)

    def SetForceZoneParameters(
        self,
        orientation: float = None,
        intensity: float = None,
        turbulence: float = None,
        turbulence_frequency: float = None,
    ):
        self._send_data(
            "SetForceZoneParameters",
            orientation is not None,
            orientation,
            intensity is not None,
            intensity,
            turbulence is not None,
            turbulence,
            turbulence_frequency is not None,
            turbulence_frequency,
        )

    def SetSolverParameters(self, gravity: list):
        self._send_data("SetSolverParameters", gravity)



from pyrfuniverse.envs import RFUniverseGymGoalWrapper
from pyrfuniverse.utils.ur5_controller import RFUniverseUR5Controller
import pyrfuniverse
import numpy as np
from gym import spaces
from gym.utils import seeding
import math
import pybullet as p
import os


class Ur5DrawerEnv(RFUniverseGymGoalWrapper):
    metadata = {"render.modes": ["human"]}

    def __init__(
        self,
        max_steps,
        reward_type="dense",
        min_open_distance=0.08,
        seed=None,
        executable_file=None,
        asset_bundle_file=None,
        assets: list = [],
    ):
        super().__init__(
            executable_file=executable_file,
            camera_channel=True,
            articulation_channel=True,
            assets=assets,
        )
        self.asset_channel.set_action(
            "InstanceObject",
            name="ur5_robotiq85",
            id=123,
        )
        self.articulation_channel.set_action(
            "SetTransform",
            id=123,
            position=[0, 0, 0],
            rotation=[0, -90, 0],
        )
        self.asset_channel.set_action(
            "InstanceObject",
            name="handled_drawer",
            id=456,
        )
        self.articulation_channel.set_action(
            "SetTransform",
            id=456,
            position=[-0.721, 0.5, -0.538],
            rotation=[-90, 180, 0],
        )
        self.camera_channel.set_action(
            "AddCamera",
            id=0,
            position=[-0.25, 0.712, -0.914],
            rotation=[26.997, -13.874, 0],
        )
        self._step()
        self.max_steps = max_steps
        self.reward_type = reward_type
        self.min_open_distance = min_open_distance
        self.asset_bundle_file = asset_bundle_file

        self.seed(seed)
        self.ik_controller = RFUniverseUR5Controller(
            robot_urdf=os.path.join(
                pyrfuniverse.assets_path,
                "RoboTube/DrawerClosing/UR5/ur5_robotiq_85.urdf",
            ),
            init_joint_positions=[90, -60, 60, 90, 90, 0],
        )
        self.t = 0
        self.goal = np.array([self.min_open_distance])
        self.action_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )
        self.eef_orn = p.getQuaternionFromEuler([math.pi, 0, 0])

    def step(self, action: np.ndarray):
        action_ctrl = action.copy()
        pos_ctrl = action_ctrl[:3] * 0.05
        curr_pos = self._get_gripper_position()
        pos_ctrl = pos_ctrl + curr_pos
        joint_positions = self.ik_controller.calculate_ik_recursive(
            pos_ctrl, eef_orn=self.eef_orn
        )

        gripper_width = self._get_gripper_width()
        gripper_width_ctrl = np.clip(gripper_width + action_ctrl[3] * 0.2, 0, 0.085)
        gripper_angle = self._compute_gripper_angle(gripper_width_ctrl)
        joint_positions.append(gripper_angle)

        self._set_ur5_robotiq85_joints(joint_positions)
        self.t += 1

        obs = self._get_obs()
        done = False
        info = {"is_success": self._check_success(obs)}
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)

        if self.t == self.max_steps:
            obs = self.reset()
            done = True

        return obs, reward, done, info

    def reset(self):
        super().reset()
        self.env.reset()
        self.ik_controller.reset()
        self.t = 0

        handle_position = np.array(self.articulation_channel.data[456]["positions"][3])
        joint_positions = self.ik_controller.calculate_ik_recursive(
            handle_position, eef_orn=self.eef_orn
        )
        gripper_angle = self._compute_gripper_angle(0.085)
        joint_positions.append(gripper_angle)
        self._set_ur5_robotiq85_joints_directly(joint_positions)

        return self._get_obs()

    def seed(self, seed=1234):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self.camera_channel.set_action("GetImages", rendering_params=[[0, 320, 240]])
        self._step()
        img = self.camera_channel.images.pop(0)

        return img

    def load_drawer(self, name, position, rotation):
        self.asset_channel.set_action(
            "LoadArticulationBody",
            filename=self.asset_bundle_file,
            name=name,
            position=position,
            rotation=rotation,
        )
        self._step()

    def destroy_drawer(self, index):
        self.articulation_channel.set_action("Destroy", index=index + 2)
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        if self.reward_type == "dense":
            return float(achieved_goal)
        else:
            overhead_angle = achieved_goal - desired_goal
            return float(overhead_angle[0] > 0)

    def add_camera(self, position, rotation):
        self.camera_channel.set_action(
            "AddCamera",
            position=position,
            rotation=rotation,
        )
        self._step()

    def get_images(self):
        self.camera_channel.set_action(
            "GetImages", rendering_params=[[0, 320, 240], [1, 320, 240]]
        )
        self._step()

        img_rgb = self.camera_channel.images.pop(0)
        img_depth = self.camera_channel.images.pop(0)

        return img_rgb, img_depth

    def _get_obs(self):
        gripper_position = self._get_gripper_position()
        gripper_width = self._get_gripper_width()

        robot_obs = np.concatenate((gripper_position, [gripper_width]))

        handle_position = np.array(self.articulation_channel.data[456]["positions"][3])
        drawer_open_distance = np.array(
            [self.articulation_channel.data[456]["joint_positions"][0]]
        )

        object_obs = np.concatenate((handle_position, drawer_open_distance))

        obs = np.concatenate((robot_obs, object_obs))

        return {
            "observation": obs.copy(),
            "achieved_goal": drawer_open_distance.copy(),
            "desired_goal": self.goal.copy(),
        }

    def _get_gripper_position(self):
        return np.array(self.articulation_channel.data[1230]["positions"][11])

    def _get_gripper_width(self):
        right_inner_finger_pos = np.array(
            self.articulation_channel.data[1230]["positions"][5]
        )
        left_inner_finger_pos = np.array(
            self.articulation_channel.data[1230]["positions"][10]
        )
        width = self._compute_distance(right_inner_finger_pos, left_inner_finger_pos)

        # The position is at the center of inner_finger, so we must get rid of the width of inner finger,
        # to get accurate gripper width
        width = width - 0.00635

        return width

    def _compute_gripper_angle(self, width):
        angle_rad = 0.715 - math.asin((width - 0.01) / 0.1143)
        angle_deg = angle_rad * 180 / math.pi

        return angle_deg

    def _set_ur5_robotiq85_joints(self, joint_positions):
        self.articulation_channel.set_action(
            "SetJointPosition", id=123, joint_positions=list(joint_positions[:6])
        )
        # self._step()

        self.articulation_channel.set_action(
            "SetJointPosition",
            id=1230,
            joint_positions=[joint_positions[6], joint_positions[6]],
        )
        self._step()

    def _set_ur5_robotiq85_joints_directly(self, joint_positions):
        self.articulation_channel.set_action(
            "SetJointPositionDirectly",
            id=123,
            joint_positions=list(joint_positions[:6]),
        )
        # self._step()

        self.articulation_channel.set_action(
            "SetJointPositionDirectly",
            id=1230,
            joint_positions=[joint_positions[6], joint_positions[6]],
        )
        self._step()

    def _compute_distance(self, point_a, point_b):
        return np.linalg.norm(point_a - point_b, axis=-1)

    def _check_success(self, obs):
        achieved_goal = obs["achieved_goal"][0]
        desired_goal = obs["desired_goal"][0]

        return (desired_goal < achieved_goal).astype(np.float32)



from pyrfuniverse.envs import RFUniverseBaseEnv
from pyrfuniverse.side_channel import SideChannel
from pyrfuniverse.side_channel import IncomingMessage
import numpy as np
import uuid
from gym import spaces
from gym.utils import seeding
from gym.utils import EzPickle
import pettingzoo
from pettingzoo.utils.conversions import parallel_wrapper_fn, to_parallel
from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector
from pettingzoo.utils import wrappers
import os
import time


class CollisionDetectionChannel(SideChannel):
    def __init__(self, possible_agents):
        super().__init__(uuid.UUID("bee25cbc-07e2-11ec-9e67-18c04d443e7d"))
        self.agent_collisions = {}
        for agent in possible_agents:
            self.agent_collisions[agent] = 0

    def on_message_received(self, msg: IncomingMessage) -> None:
        for i in range(len(self.agent_collisions.items())):
            agent_name = msg.read_string()
            num_collisions = msg.read_int32()

            if agent_name == "":
                continue
            if agent_name[-7:] == "(Clone)":
                agent_name = agent_name[:-7]
            self.agent_collisions[agent_name] += num_collisions

        # print(self.agent_collisions)

    def get_agent_collisions(self):
        agent_collisions = self.agent_collisions.copy()
        self.clear()

        return agent_collisions

    def clear(self):
        for key, value in self.agent_collisions.items():
            self.agent_collisions[key] = 0


def env(**kwargs):
    env = raw_env(**kwargs)
    # if env.continuous:
    # env = wrappers.ClipOutOfBoundsWrapper(env)
    # else:
    # env = wrappers.AssertOutOfBoundsWrapper(env)
    env = wrappers.OrderEnforcingWrapper(env)
    return env


parallel_env = parallel_wrapper_fn(env)


class raw_env(AECEnv, RFUniverseBaseEnv, EzPickle):
    metadata = {"render.modes": ["human"], "name": "navigation_v1"}
    reward_range = (-float("inf"), float("inf"))

    def __init__(
        self,
        num_agents,
        asset_bundle_file,
        log_dir,
        reset_on_collision=False,
        max_episode_length=100,
        executable_file=None,
        log_monitor=True,
        strength=1000,
        collision_multiplier=5,
    ):
        """
        The init method takes in environment arguments and
         should define the following attributes:
        - possible_agents
        - action_spaces
        - observation_spaces

        These attributes should not be changed after initialization.
        """
        EzPickle.__init__(
            self,
            num_agents,
            asset_bundle_file,
            log_dir,
            reset_on_collision,
            max_episode_length,
            executable_file,
            log_monitor,
        )

        self.possible_agents = ["agent_" + str(r) for r in range(num_agents)]
        self.collision_detection_channel = CollisionDetectionChannel(
            self.possible_agents
        )
        RFUniverseBaseEnv.__init__(
            self,
            executable_file=executable_file,
            custom_channels=[self.collision_detection_channel],
            rigidbody_channel=True,
        )

        self.asset_bundle_file = asset_bundle_file
        self.reset_on_collision = reset_on_collision
        self.max_episode_length = max_episode_length
        self.log_dir = log_dir
        self.log_monitor = log_monitor
        self.collision_multiplier = collision_multiplier
        self.strength = strength

        if self.log_monitor:
            os.makedirs(self.log_dir, exist_ok=True)
            self.monitor_id = 0
            while os.path.exists(
                os.path.join(self.log_dir, "{}.monitor.csv".format(self.monitor_id))
            ):
                self.monitor_id += 1
            with open(
                os.path.join(self.log_dir, "{}.monitor.csv".format(self.monitor_id)),
                "a+",
            ) as f:
                f.write("r,l,t\n")
        self.counter = 0
        self.total_reward = 0
        self.start_time = time.time()

        # Fixed parameters
        self.agent_name = "NavRobot"
        self.y_offset = 0.15
        self.reset_agent_min_distance = 1.1
        self.world_range_low = np.array([-4, self.y_offset, -4])
        self.world_range_high = np.array([4, self.y_offset, 4])

        self.agent_name_mapping = dict(
            zip(self.possible_agents, list(range(len(self.possible_agents))))
        )
        # print(self.agent_name_mapping)

        self.action_spaces = {
            agent: spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
            for agent in self.possible_agents
        }
        self.observation_spaces = {
            agent: spaces.Box(
                low=-np.inf, high=np.inf, shape=(num_agents * 4,), dtype=np.float32
            )
            for agent in self.possible_agents
        }

        self.seed()
        self._env_setup(num_agents)
        self.t = 0

    def observe(self, agent):
        """
        Observe should return the observation of the specified agent. This function
        should return a sane observation (though not necessarily the most up to date possible)
        at any time after reset() is called.
        """
        obs = self._get_obs_for_agent(agent)
        return obs

    def reset(self):
        """
        Reset needs to initialize the following attributes
        - agents
        - rewards
        - _cumulative_rewards
        - dones
        - infos
        - agent_selection
        And must set up the environment so that render(), step(), and observe()
        can be called without issues.

        Here it sets up the state dictionary which is used by step() and the observations dictionary which is used by step() and observe()
        """
        self.agents = self.possible_agents[:]
        self.rewards = {agent: 0 for agent in self.agents}
        self._cumulative_rewards = {agent: 0 for agent in self.agents}
        self.dones = {agent: False for agent in self.agents}
        self.infos = {agent: {} for agent in self.agents}

        self.t = 0
        self._env_reset()
        self.collision_detection_channel.clear()
        """
        Our agent_selector utility allows easy cyclic stepping through the agents list.
        """
        self._agent_selector = agent_selector(self.agents)
        self.agent_selection = self._agent_selector.next()

    def step(self, action):
        """
        step(action) takes in an action for the current agent (specified by
        agent_selection) and needs to update
        - rewards
        - _cumulative_rewards (accumulating the rewards)
        - dones
        - infos
        - agent_selection (to the next agent)
        And any internal state used by observe() or render()
        """
        if self.dones[self.agent_selection]:
            # handles stepping an agent which is already done
            # accepts a None action for the one agent, and moves the agent_selection to
            # the next done agent,  or if there are no more done agents, to the next live agent
            return self._was_done_step(action)

        agent = self.agent_selection

        # the agent which stepped last had its _cumulative_rewards accounted for
        # (because it was returned by last()), so the _cumulative_rewards for this
        # agent should start again at 0
        self._cumulative_rewards[agent] = 0

        self._set_agent_force(
            self.agent_name_mapping[agent],
            np.array([action[0], 0, action[1]]) * self.strength,
        )

        done = False
        # collect reward if it is the last agent to act
        if self._agent_selector.is_last():
            agent_collisions = self.collision_detection_channel.get_agent_collisions()
            # print(agent_collisions)
            # rewards for all agents are placed in the .rewards dictionary
            for agent_name in self.agents:
                velocity = self._get_velocity_for_agent(agent_name)
                reward = (
                    velocity - agent_collisions[agent_name] * self.collision_multiplier
                )
                self.rewards[agent_name] = reward

            self.t += 1
            # The dones dictionary must be updated for all players.
            done = self.t >= self.max_episode_length
            self.dones = {agent: done for agent in self.agents}
        else:
            self._clear_rewards()

        # selects the next agent.
        self.agent_selection = self._agent_selector.next()
        # Adds .rewards to ._cumulative_rewards
        self._accumulate_rewards()

        if self.log_monitor and done:
            self._log()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def close(self):
        RFUniverseBaseEnv.close(self)

    def _env_setup(self, num_agents):
        for i in range(num_agents):
            agent_position = self.np_random.uniform(
                self.world_range_low, self.world_range_high
            )
            self.asset_channel.set_action(
                "LoadRigidbodyWithName",
                filename=self.asset_bundle_file,
                name=self.agent_name,
                replace_name=self.possible_agents[i],
                position=list(agent_position),
            )
            self._step()

    def _env_reset(self):
        positions = []
        for i in range(self.num_agents):
            while True:
                agent_position = self.np_random.uniform(
                    self.world_range_low, self.world_range_high
                )
                if self._check_reset_position_legality(positions, agent_position):
                    break
            positions.append(agent_position.copy())
            self._set_agent_position(i, agent_position)

    def _set_agent_position(self, index, position):
        self.rigidbody_channel.set_action(
            "SetTransform", index=index, position=list(position), rotation=[0, 0, 0]
        )
        self._step()

    def _set_agent_force(self, index, force):
        self.rigidbody_channel.set_action("AddForce", index=index, force=force)
        self._step()

    def _check_reset_position_legality(self, positions, agent_pos):
        for position in positions:
            if np.linalg.norm(position - agent_pos) < self.reset_agent_min_distance:
                return False

        return True

    def _get_obs_for_agent(self, agent):
        agent_id = self.agent_name_mapping[agent]
        self_pos = self.rigidbody_channel.data[agent_id]["position"]
        self_vel = self.rigidbody_channel.data[agent_id]["velocity"]
        self_obs = np.array([self_pos[0], self_pos[2], self_vel[0], self_vel[2]])

        other_all_obs = np.array([])
        for i in range(self.num_agents):
            if i == agent_id:  # Ignore itself
                continue
            other_pos = self.rigidbody_channel.data[i]["position"]
            other_vel = self.rigidbody_channel.data[i]["velocity"]
            other_obs = (
                np.array([other_pos[0], other_pos[2], other_vel[0], other_vel[2]])
                - self_obs
            )
            other_all_obs = np.concatenate((other_all_obs, other_obs))

        return np.concatenate((self_obs, other_all_obs)).copy()

    def _get_velocity_for_agent(self, agent):
        agent_id = self.agent_name_mapping[agent]
        self_vel = self.rigidbody_channel.data[agent_id]["velocity"]

        return np.linalg.norm(np.array([self_vel[0], self_vel[2]]))

    def _log(self):
        self.counter += self.max_episode_length
        total_reward = 0
        for agent_name in self.agents:
            total_reward += self.rewards[agent_name]
        with open(
            os.path.join(self.log_dir, "{}.monitor.csv".format(self.monitor_id)), "a+"
        ) as f:
            f.write(
                "{},{},{}\n".format(
                    total_reward, self.counter, time.time() - self.start_time
                )
            )



from enum import Enum
import pyrfuniverse.attributes as attr


class LightType(Enum):
    """
    The type of light, keeping same name with LightType (https://docs.unity3d.com/ScriptReference/LightType.html) in Unity.
    """

    Spot = 0
    Directional = 1
    Point = 2
    Area = 3  # unused
    Disc = 4  # unused


class LightShadow(Enum):
    """
    The type of shadow, keeping same name with LightShadows (https://docs.unity3d.com/ScriptReference/LightShadows.html) in Unity.
    """

    NoneShadow = 0
    Hard = 1
    Soft = 2


class LightAttr(attr.BaseAttr):
    """
    Light attribute class.
    """

    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.
        """
        super().parse_message(data)

    def SetColor(self, color: list):
        """
        Set the color of light.

        Args:
            color: A list of length 3, representing the R, G and B channel, in range [0, 1].
        """
        assert color is not None and len(color) == 3, "position length must be 3"
        color = [float(i) for i in color]

        self._send_data("SetColor", color)

    def SetType(self, light_type: LightType):
        """
        Set the type of light.

        Args:
            light_type: LightType, the type of light.
        """
        self._send_data("SetType", light_type.value)

    def SetShadow(self, light_shadow: LightShadow):
        """
        Set the type of shadow.

        Args:
            light_shadow: LightShadow, the type of the shadow.
        """
        self._send_data("SetShadow", light_shadow.value)

    def SetIntensity(self, light_intensity: float):
        """
        Set the intensity of light.

        Args:
            light_intensity: Float, the intensity of light.
        """
        self._send_data("SetIntensity", float(light_intensity))

    def SetRange(self, light_range: float):
        """
        Set the range of light. (Only available when the LightType is `LightType.Spot` or `LightType.Point`)

        Args:
            light_range: Float, the range of light.
        """
        self._send_data("SetRange", float(light_range))

    def SetSpotAngle(self, spot_angle: float):
        """
        Set the angle of light. (Only available when the LightType is `LightType.Spot`)

        Args:
            spot_angle: Float, the angle of light.
        """
        self._send_data("SetSpotAngle", float(spot_angle))



from pyrfuniverse.envs.gym_goal_wrapper_env import RFUniverseGymGoalWrapper
import numpy as np
from gym import spaces
from gym.utils import seeding
import copy


class FrankaRoboticsEnv(RFUniverseGymGoalWrapper):
    metadata = {"render.modes": ["human"]}
    # height_offset is just the object width / 2, which is the object's stable height value.
    height_offset = 0.025

    def __init__(
        self,
        max_episode_length,
        reward_type,
        tolerance,
        load_object,
        target_in_air,
        block_gripper,
        target_xz_range,
        target_y_range,
        object_xz_range,
        seed=1234,
        executable_file=None,
        scene_file=None,
        asset_bundle_file=None,
        assets: list = [],
    ):
        super().__init__(executable_file, scene_file, assets=assets)
        self.max_steps = max_episode_length
        self.reward_type = reward_type
        self.tolerance = tolerance
        self.load_object = load_object
        self.target_in_air = target_in_air
        self.block_gripper = block_gripper
        self.goal_range_low = np.array(
            [-target_xz_range, self.height_offset, -target_xz_range]
        )
        self.goal_range_high = np.array(
            [target_xz_range, target_y_range, target_xz_range]
        )
        self.object_range_low = np.array(
            [-object_xz_range, self.height_offset, -object_xz_range]
        )
        self.object_range_high = np.array(
            [object_xz_range, self.height_offset, object_xz_range]
        )
        self.asset_bundle_file = asset_bundle_file

        self.seed(seed)
        self._env_setup()
        self.init_pos = [0.15, 0.3, 0]
        self.t = 0
        self.goal = self._sample_goal()
        self.action_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)

        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )

    def step(self, action: np.ndarray):
        """
        Params:
            action: 4-d numpy array.
        """
        # print(action)
        pos_ctrl = action[:3] * 0.05
        curr_pos = np.array(self.attrs[9658740].data["positions"][3])
        # print(curr_pos)
        pos_ctrl = curr_pos + pos_ctrl

        # print(pos_ctrl)
        # self.instance_channel.set_action(
        #     "IKTargetDoMove",
        #     id=965874,
        #     position=[pos_ctrl[0], pos_ctrl[1], pos_ctrl[2]],
        #     duration=0.1,
        #     speed_based=True
        # )
        self.attrs[965874].IKTargetDoMove(
            position=[pos_ctrl[0], pos_ctrl[1], pos_ctrl[2]],
            duration=0.1,
            speed_based=True,
        )
        if self.block_gripper:
            self._set_gripper_width(0)
        else:
            curr_gripper_width = self._get_gripper_width()
            target_gripper_width = curr_gripper_width + action[3] * 0.2
            target_gripper_width = np.clip(target_gripper_width, 0, 0.08)
            self._set_gripper_width(target_gripper_width)

        # self._set_franka_joints(np.array(joint_positions))
        self._step()
        self.t += 1

        obs = self._get_obs()
        done = False
        info = {"is_success": self._check_success(obs)}
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)

        if self.t == self.max_steps:
            done = True

        return obs, reward, done, info

    def reset(self):
        super().reset()
        self.t = 0
        self.goal = self._sample_goal()
        object_pos = None

        if self.load_object:
            object_pos = self._reset_object()

        # self.instance_channel.set_action(
        #     'SetTransform',
        #     id=0,
        #     position=list(self.goal)
        # )
        # self._step()

        # self.ik_controller.reset()

        if self.load_object and self.target_in_air:
            self.attrs[965874].IKTargetDoMove(
                position=[self.init_pos[0], self.init_pos[1], self.init_pos[2]],
                duration=0,
                speed_based=False,
            )
            self._step()
            self.attrs[9658740].SetJointPositionDirectly(joint_positions=[0.04, 0.04])
            self._step()

        self._step()

        return self._get_obs()

    def seed(self, seed=1234):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        distance = self._compute_goal_distance(achieved_goal, desired_goal)
        if self.reward_type == "sparse":
            return -(distance > self.tolerance).astype(np.float32)
        else:
            return -distance

    def _get_obs(self):
        gripper_position = np.array(self.attrs[9658740].data["positions"][3])
        gripper_velocity = np.array(self.attrs[9658740].data["velocities"][3])
        gripper_width = self._get_gripper_width()
        # gripper_joint_position = np.array(self.articulation_channel.data[1]['joint_positions'])
        # gripper_joint_velocity = np.array(self.articulation_channel.data[1]['joint_velocities'])

        panda_obs = np.concatenate(
            (gripper_position, gripper_velocity, [gripper_width])
        )

        if self.load_object:
            object_position = np.array(self.attrs[0].data["position"])
            object_rotation = np.array(self.attrs[0].data["rotation"])
            object_velocity = np.array(self.attrs[0].data["velocity"])
            object_angular_vel = np.array(self.attrs[0].data["angular_vel"])
            # object_rel_pos = object_position - gripper_position
            # object_velocity = object_velocity - gripper_velocity
            achieved_goal = object_position.copy()
        else:
            # object_position = object_rotation = object_velocity = object_angular_vel = object_rel_pos = np.zeros(0)
            object_position = (
                object_rotation
            ) = object_velocity = object_angular_vel = np.zeros(0)
            achieved_goal = gripper_position.copy()

        object_obs = np.concatenate(
            (object_position, object_rotation, object_velocity, object_angular_vel)
        )

        # obs = np.concatenate(
        #     [gripper_position, object_position, object_rel_pos, gripper_joint_position, object_rotation,
        #     object_velocity, object_angular_vel, gripper_velocity, gripper_joint_velocity]
        # )
        obs = np.concatenate((panda_obs, object_obs))

        return {
            "observation": obs.copy(),
            "achieved_goal": achieved_goal.copy(),
            "desired_goal": self.goal.copy(),
        }

    def _env_setup(self):
        if self.load_object:
            self.InstanceObject(name="Rigidbody_Box", id=0)
            self.attrs[0].SetTransform(
                position=[0, self.height_offset, 0], scale=[0.05, 0.05, 0.05]
            )

        self.attrs[965874].SetIKTargetOffset(position=[0, 0.105, 0])
        self.attrs[965874].IKTargetDoRotate(
            rotation=[0, 45, 180], duration=0, speed_based=False
        )
        self._step()

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min < max, "Min value is {}, while max value is {}.".format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float

    def _set_gripper_width(self, w: float):
        w = w / 2
        self.attrs[9658740].SetJointPosition(joint_positions=[w, w])

    def _get_gripper_width(self) -> float:
        gripper_joint_positions = copy.deepcopy(
            self.attrs[9658740].data["joint_positions"]
        )
        return -1 * (gripper_joint_positions[0] + gripper_joint_positions[1])

    def _check_success(self, obs):
        achieved_goal = obs["achieved_goal"]
        desired_goal = obs["desired_goal"]
        distance = self._compute_goal_distance(achieved_goal, desired_goal)

        return (distance < self.tolerance).astype(np.float32)

    def _compute_goal_distance(self, goal_a, goal_b):
        assert goal_a.shape == goal_b.shape
        return np.linalg.norm(goal_a - goal_b, axis=-1)

    def _sample_goal(self):
        if self.load_object:
            goal = np.array([0.0, 0.4, 0.0])
            noise = self.np_random.uniform(self.goal_range_low, self.goal_range_high)
            if self.target_in_air and self.np_random.random() < 0.3:
                noise[1] = self.height_offset
            goal += noise
        else:
            goal = self.np_random.uniform(self.goal_range_low, self.goal_range_high)

        return goal.copy()

    def _reset_object(self):
        object_pos = self.np_random.uniform(
            self.object_range_low, self.object_range_high
        )
        self.attrs[0].SetTransform(position=list(object_pos), rotation=[0, 0, 0])
        self._step()
        return object_pos.copy()



from pyrfuniverse.attributes.base_attr import BaseAttr
from pyrfuniverse.attributes.camera_attr import CameraAttr
from pyrfuniverse.attributes.gameobject_attr import GameObjectAttr
from pyrfuniverse.attributes.light_attr import LightAttr
from pyrfuniverse.attributes.activelightsensor_attr import ActiveLightSensorAttr
from pyrfuniverse.attributes.collider_attr import ColliderAttr
from pyrfuniverse.attributes.controller_attr import ControllerAttr
from pyrfuniverse.attributes.rigidbody_attr import RigidbodyAttr
from pyrfuniverse.attributes.cloth_attr import ClothAttr
from pyrfuniverse.attributes.softbody_attr import SoftbodyAttr
from pyrfuniverse.attributes.humanbody_attr import HumanbodyAttr
from pyrfuniverse.attributes.graspsim_attr import GraspSimAttr
from pyrfuniverse.attributes.digit_attr import DigitAttr
from pyrfuniverse.attributes.pointcloud_attr import PointCloudAttr
from pyrfuniverse.attributes.fallingcloth_attr import FallingClothAttr
from pyrfuniverse.attributes.custom_attr import CustomAttr

try:
    from pyrfuniverse.attributes.omplmanager_attr import OmplManagerAttr
except ImportError:
    pass

__all__ = [
    "base_attr",
    "camera_attr",
    "activelightsensor_attr",
    "collider_attr",
    "controller_attr",
    "gameobject_attr",
    "rigidbody_attr",
    "cloth_attr",
    "softbody_attr",
    "humanbody_attr",
    "graspsim_attr",
    "digit_attr",
    "pointcloud_attr",
    "fallingcloth_attr",
    "custom_attr",
    "omplmanager_attr",
]



from pyrfuniverse.envs.gym_goal_wrapper_env import RFUniverseGymGoalWrapper
import numpy as np
from gym import spaces
from gym.utils import seeding
import math
from pyrfuniverse.utils.interpolate_utils import sine_interpolate


class FlexivCuttingEnv(RFUniverseGymGoalWrapper):
    metadata = {"render.modes": ["human", "rgb_array"]}
    object2id = {
        "flexiv": 621325,
        "ag95": 6213250,
        "target": 9413,
        "knife": 35452,
        "goal_knife": 35451,
    }

    def __init__(
        self,
        max_steps=200,
        knife_pos_min=(-0.5, 0.05, -0.07),
        knife_pos_max=(-0.6, 0.07, 0.06),
        knife_rot_min=(90, 0, -30),
        knife_rot_max=(90, 0, 30),
        offset_tolerance=0.04,
        angle_tolerance=5,
        reward_type="dense",
        consider_rotation=True,
        executable_file=None,
        assets: list = [],
    ):
        super().__init__(executable_file=executable_file, assets=assets)
        self._reload_scene()
        self.scale = 5
        self.fixed_delta_time = 0.02
        self.max_steps = max_steps
        self.knife_pos_min = np.array(knife_pos_min)
        self.knife_pos_max = np.array(knife_pos_max)
        self.knife_rot_min = np.array(knife_rot_min)
        self.knife_rot_max = np.array(knife_rot_max)
        self.offset_tolerance = offset_tolerance
        self.angle_tolerance = angle_tolerance
        self.consider_rotation = consider_rotation
        self.reward_type = reward_type

        self.seed()
        self.init_joint_positions = [65, 0, 34, -90, 0, 0, -78]
        self.init_eef_euler = [180, 0, 0]
        self.eef_euler = self.init_eef_euler.copy()
        self.goal = self._sample_goal()
        self.t = 0
        self._init_env_()
        self.action_space = spaces.Box(
            low=-1,
            high=1,
            shape=(4,) if self.consider_rotation else (3,),
            dtype=np.float32,
        )
        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )

    def step(self, action: np.ndarray):
        assert (action.shape == (4,) and self.consider_rotation) or (
            action.shape == (3,) and not self.consider_rotation
        )
        current_pos = self._get_eef_pos()
        delta_pos = action[:3] * 0.05
        target_pos = current_pos + delta_pos

        if self.consider_rotation:
            delta_euler = action[3] * 5
            target_euler = self.eef_euler + np.array([0, delta_euler, 0])
            self.eef_euler = np.clip(
                target_euler,
                a_min=np.array([140, -40, -40]),
                a_max=np.array([220, 40, 40]),
            )

        self._set_flexiv(target_pos, self.eef_euler)

        self.t += 1
        obs = self._get_obs()
        info = {
            "is_success": self._check_success(obs["achieved_goal"], obs["desired_goal"])
        }
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)
        done = False

        if self.t == self.max_steps:
            done = True

        return obs, reward, done, info

    def reset(self):
        super().reset()
        self.t = 0
        self.eef_euler = self.init_eef_euler.copy()
        self._reload_scene()
        self._init_env_()
        self.goal = self._sample_goal()
        self._set_goal_knife()

        return self._get_obs()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        assert len(achieved_goal.shape) <= 2
        if len(achieved_goal.shape) < 2:
            offset = self._distance(achieved_goal[:3], desired_goal[:3])
            if self.consider_rotation:
                angle = self._angle_from_quaternions(
                    achieved_goal[3:7], desired_goal[3:7]
                )
            else:
                angle = np.array(0)
        else:
            offset = self._distance(achieved_goal[:, :3], desired_goal[:, :3])
            if self.consider_rotation:
                angle = self._angle_from_quaternions(
                    achieved_goal[:, 3:7], desired_goal[:, 3:7]
                )
            else:
                angle = np.array(0)

        if self.reward_type == "dense":
            return -(offset + angle).astype(np.float32)
        else:
            return -(offset > self.offset_tolerance).astype(np.float32) - (
                angle > self.angle_tolerance
            ).astype(np.float32)

    def heuristic(self):
        target_pos = self._get_target_pos()
        # print(target_pos, self._get_eef_pos())
        target_rot = self._get_target_rot()
        self._set_flexiv_directly(target_pos, target_rot)
        # self._step()

        achieved_quat = self._get_knife_quat()
        desired_rot = self._get_goal_knife_quat()
        # print(self.instance_channel.data[self.object2id['flexiv']]['joint_positions'])
        # print(self._angle_from_quaternions(achieved_quat, desired_rot))

    def naive_cutting(self, num_steps=100):
        curr_pos = self._get_eef_pos()
        delta_pos = np.array([0, 0.3, 0]) / self.scale
        for i in range(num_steps):
            target_pos = curr_pos - delta_pos / num_steps * (i + 1)
            self._set_flexiv_directly(target_pos, self.eef_euler)
            self._step()

    def naive_cutting_old_version(self, num_steps=100):
        eef_positions = []
        curr_target_pos = self._get_target_pos()
        delta_pos = np.array([0, 0.3, 0]) / self.scale
        for i in range(num_steps):
            target_pos = curr_target_pos - delta_pos / num_steps * (i + 1)
            self._set_target_pos(target_pos)
            self.heuristic()
            if i % 10 == 0:
                eef_positions.append(self.get_flexiv_eef_pos())

        eef_positions.append(self.get_flexiv_eef_pos())

        return eef_positions

    def replay(self, eef_positions, num_steps=20):
        times = [2.82, 3.62, 2.78, 3.15, 1.82]
        time_counter = 0
        for i in range(eef_positions.shape[0]):
            if 1 <= i <= 6:
                continue
            target_position = eef_positions[i]
            curr_position = self._get_eef_pos()
            if i > 1 and time_counter < len(times):
                this_num_steps = int(times[time_counter] / 0.05)
                time_counter += 1
            else:
                this_num_steps = num_steps
            trajectories = sine_interpolate(
                curr_position, target_position, this_num_steps
            )
            for traj in trajectories:
                self._set_flexiv(traj, self.eef_euler)

            print("Demo " + str(i + 1) + " Completed! " + str(this_num_steps))

    def move_to_joint_position(self):
        init_joint_positions = (
            np.array(
                [
                    2.56255080e-05,
                    -6.98036730e-01,
                    -2.39185442e-06,
                    -1.57088447e00,
                    2.91160322e-05,
                    6.98098183e-01,
                    -6.93326874e-05,
                ]
            )
            / math.pi
            * 180
        )
        # joint_positions[3] need to be inverse

        self.attrs[self.object2id["flexiv"]].EnabledNativeIK(False)
        self._step()
        self.attrs[self.object2id["flexiv"]].SetJointPositionDirectly(
            init_joint_positions
        )
        self._step()

    def _get_obs(self):
        eef_pos = self._get_eef_pos()
        eef_quat = self._get_eef_quat()
        eef_velocity = self._get_eef_velocity()
        robot_obs = np.concatenate((eef_pos, eef_quat, eef_velocity))

        knife_pos = self._get_knife_pos()
        knife_quat = self._get_knife_quat()
        knife_obs = np.concatenate((knife_pos, knife_quat))

        goal_knife_obs = np.concatenate((self.goal[:3], self._get_goal_knife_quat()))

        if self.consider_rotation:
            achieved_goal = knife_obs.copy()
            desired_goal = goal_knife_obs.copy()
        else:
            achieved_goal = knife_pos.copy()
            desired_goal = self.goal.copy()

        return {
            "observation": np.concatenate((robot_obs, knife_obs, goal_knife_obs)),
            "achieved_goal": achieved_goal.copy(),
            "desired_goal": desired_goal.copy(),
        }

    def _init_env_(self):
        self._init_flexiv()
        self._init_ag95()

    def _reload_scene(self):
        self.SwitchSceneAsync("FlexivSOFA", True)
        self._step()

    def _init_flexiv_target_offset(self, offset):
        self.attrs[self.object2id["flexiv"]].SetIKTargetOffset(position=offset)
        self._step()

    def _init_flexiv(self):
        self.attrs[self.object2id["flexiv"]].EnabledNativeIK(False)
        self._step()
        self.attrs[self.object2id["flexiv"]].SetJointPositionDirectly(
            joint_positions=self.init_joint_positions
        )
        self._step()
        self.attrs[self.object2id["flexiv"]].EnabledNativeIK(True)
        self._step()

        self._init_flexiv_target_offset([0, 0, 0])
        self._step()

        self._set_flexiv_directly(
            position=np.array([-0.5, 0.5, 0]), rotation=[180, 0, 0]
        )
        self._step()

        self._init_flexiv_target_offset([0, 0.26, 0])
        self._step()

    def _init_ag95(self):
        self.attrs[self.object2id["ag95"]].SetJointPositionDirectly(
            joint_positions=[55, 55]
        )
        self._step()

    def _set_flexiv(self, position, rotation=None):
        self.attrs[self.object2id["flexiv"]].IKTargetDoMove(
            position=position * self.scale,
            duration=self.fixed_delta_time,
            speed_based=False,
        )
        if rotation is not None:
            self.attrs[self.object2id["flexiv"]].IKTargetDoRotate(
                rotation=rotation, duration=self.fixed_delta_time, speed_based=False
            )
        self._step()

    def _set_target_pos(self, position):
        self.attrs[self.object2id["target"]].SetTransform(
            position=list(np.array(position) * self.scale)
        )
        self._step()

    def _get_eef_pos(self):
        return (
            np.array(self.attrs[self.object2id["ag95"]].data["positions"][9])
            / self.scale
        )

    def _get_eef_rot(self):
        return (
            np.array(self.attrs[self.object2id["ag95"]].data["rotations"][9])
            / 180
            * math.pi
        )

    def _get_eef_quat(self):
        return np.array(self.attrs[self.object2id["ag95"]].data["quaternions"][9])

    def _get_eef_velocity(self):
        return (
            np.array(self.attrs[self.object2id["ag95"]].data["velocities"][9])
            / self.scale
        )

    def get_flexiv_eef_pos(self):
        return (
            np.array(self.attrs[self.object2id["ag95"]].data["positions"][0])
            / self.scale
        )

    def _get_target_pos(self):
        return (
            np.array(self.attrs[self.object2id["target"]].data["position"]) / self.scale
        )

    def _get_target_rot(self):
        return np.array(self.attrs[self.object2id["target"]].data["rotation"])

    def _get_knife_pos(self):
        return np.array(
            np.array(self.attrs[self.object2id["knife"]].data["position"]) / self.scale
        )

    def _get_knife_quat(self):
        return np.array(self.attrs[self.object2id["knife"]].data["quaternion"])

    def _get_goal_knife_quat(self):
        return np.array(self.attrs[self.object2id["goal_knife"]].data["quaternion"])

    def _set_flexiv_directly(self, position, rotation=None):
        self.attrs[self.object2id["flexiv"]].IKTargetDoMove(
            position=position * self.scale, duration=0, speed_based=False
        )
        if rotation is not None:
            self.attrs[self.object2id["flexiv"]].IKTargetDoRotate(
                rotation=rotation, duration=0, speed_based=False
            )
        self._step()

    def _set_goal_knife(self):
        if self.consider_rotation:
            self.attrs[self.object2id["goal_knife"]].SetTransform(
                position=list(self.goal[:3] * self.scale),
                rotation=list(self.goal[3:6] * 180 / math.pi),
            )
        else:
            self.attrs[self.object2id["goal_knife"]].SetTransform(
                position=list(self.goal[:3] * self.scale)
            )
        self._step()

    def _sample_goal(self):
        sampled_pos = self.np_random.uniform(self.knife_pos_min, self.knife_pos_max)
        if self.consider_rotation:
            sampled_rot = (
                self.np_random.uniform(self.knife_rot_min, self.knife_rot_max)
                / 180
                * math.pi
            )
        else:
            sampled_rot = np.array([])

        # Fixed Demo
        # sampled_pos = np.array([-0.54, 0.05, 0.01])
        # sampled_rot = np.array([90, 0, 0]) / 180 * math.pi

        return np.concatenate((sampled_pos, sampled_rot))

    def _distance(self, pos1, pos2):
        assert pos2.shape == pos1.shape
        return np.linalg.norm(pos1 - pos2, axis=-1)

    def _angle_from_quaternions(self, quat1, quat2):
        num = np.clip(abs(np.sum(quat1 * quat2, axis=-1)), a_max=1, a_min=-1)
        return np.arccos(num) * 2 * 180 / math.pi

    def _check_success(self, achieved_goal, desired_goal):
        offset = self._distance(achieved_goal[:3], desired_goal[:3])

        if self.consider_rotation:
            angle = self._angle_from_quaternions(achieved_goal[3:7], desired_goal[3:7])
        else:
            angle = 0

        return float(offset < self.offset_tolerance and angle < self.angle_tolerance)



from typing import List
import struct


class IncomingMessage:
    """
    Utility class for reading the message written to a SideChannel.
    Values must be read in the order they were written.
    """

    def __init__(self, buffer: bytes, offset: int = 0):
        """
        Create a new IncomingMessage from the bytes.
        """
        self.buffer = buffer
        self.offset = offset

    def read_bool(self, default_value: bool = False) -> bool:
        """
        Read a boolean value from the message buffer.
        :param default_value: Default value to use if the end of the message is reached.
        :return: The value read from the message, or the default value if the end was reached.
        """
        if self._at_end_of_buffer():
            return default_value

        self.offset += 1
        return bool(
            int.from_bytes(
                self.buffer[self.offset - 1 : self.offset], byteorder="little"
            )
        )

    def read_int32(self, default_value: int = 0) -> int:
        """
        Read an integer value from the message buffer.
        :param default_value: Default value to use if the end of the message is reached.
        :return: The value read from the message, or the default value if the end was reached.
        """
        if self._at_end_of_buffer():
            return default_value

        self.offset += 4
        return int.from_bytes(
            self.buffer[self.offset - 4 : self.offset], byteorder="little"
        )

    def read_float32(self, default_value: float = 0.0) -> float:
        """
        Read a float value from the message buffer.
        :param default_value: Default value to use if the end of the message is reached.
        :return: The value read from the message, or the default value if the end was reached.
        """
        if self._at_end_of_buffer():
            return default_value

        self.offset += 4
        return struct.unpack("f", self.buffer[self.offset - 4 : self.offset])[0]

    def read_float32_list(self, default_value: List[float] = None) -> List[float]:
        """
        Read a list of float values from the message buffer.
        :param default_value: Default value to use if the end of the message is reached.
        :return: The value read from the message, or the default value if the end was reached.
        """
        if self._at_end_of_buffer():
            return [] if default_value is None else default_value

        list_len = self.read_int32()
        output = []
        for _ in range(list_len):
            output.append(self.read_float32())
        return output

    def read_string(self, default_value: str = "") -> str:
        """
        Read a string value from the message buffer.
        :param default_value: Default value to use if the end of the message is reached.
        :return: The value read from the message, or the default value if the end was reached.
        """
        if self._at_end_of_buffer():
            return default_value

        count = self.read_int32()
        self.offset += count
        return self.buffer[self.offset - count : self.offset].decode("utf-8")

    def _at_end_of_buffer(self) -> bool:
        return self.offset >= len(self.buffer)



from pyrfuniverse.envs.gym_goal_wrapper_env import RFUniverseGymGoalWrapper

# from pyrfuniverse.utils import RFUniverseController
import numpy as np
from gym import spaces
from gym.utils import seeding
import copy


class FrankaClothFoldEnv(RFUniverseGymGoalWrapper):
    metadata = {"render.modes": ["human"]}
    height_offset = 0.01

    def __init__(
        self,
        asset_bundle_file,
        executable_file=None,
        reward_type="sparse",
        tolerance=0.05,
        object_xz_range=0.055,
        object_rotation_range=180,
    ):
        super().__init__(
            executable_file=executable_file,
            articulation_channel=True,
            obi_cloth_with_grasping_channel=True,
        )
        self.reward_type = reward_type
        self.tolerance = tolerance
        self.object_xz_range = object_xz_range
        self.object_rotation_range = object_rotation_range
        self.object_range_low = np.array(
            [-object_xz_range, self.height_offset, -object_xz_range]
        )
        self.object_range_high = np.array(
            [object_xz_range, self.height_offset, object_xz_range]
        )
        self.asset_bundle_file = asset_bundle_file
        self.ik_controller = RFUniverseController(
            "franka", base_pos=np.array([-0.6, 0, 0])
        )
        self.t = 0

        self.has_loaded_object = False
        self.goal = None
        self.seed()
        self._load_object()

        self.action_space = spaces.Box(low=-1, high=1, shape=(4,), dtype=np.float32)
        obs = self._get_obs()
        self.observation_space = spaces.Dict(
            {
                "observation": spaces.Box(
                    -np.inf, np.inf, shape=obs["observation"].shape, dtype=np.float32
                ),
                "desired_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["desired_goal"].shape, dtype=np.float32
                ),
                "achieved_goal": spaces.Box(
                    -np.inf, np.inf, shape=obs["achieved_goal"].shape, dtype=np.float32
                ),
            }
        )

    def step(self, action: np.ndarray):
        pos_ctrl = action[:3] * 0.05
        curr_pos = np.array(self.articulation_channel.data[1]["positions"][3])
        pos_ctrl = curr_pos + pos_ctrl
        joint_positions = self.ik_controller.calculate_ik_recursive(pos_ctrl)

        curr_gripper_width = self._get_gripper_width()
        target_gripper_width = curr_gripper_width + action[3] * 0.2
        target_gripper_width = np.clip(target_gripper_width, 0, 0.08)
        joint_positions.append(target_gripper_width)

        self._set_franka_joints(np.array(joint_positions))
        self.t += 1

        obs = self._get_obs()
        done = False
        info = {"is_success": self._check_success(obs)}
        reward = self.compute_reward(obs["achieved_goal"], obs["desired_goal"], info)

        return obs, reward, done, info

    def reset(self):
        super().reset()
        self._destroy_object()
        self.env.reset()
        self.ik_controller.reset()
        self.t = 0
        self._load_object()

        grasp_position = self._get_grasp_position()
        joint_positions = self.ik_controller.calculate_ik_recursive(grasp_position)
        self.articulation_channel.set_action(
            "SetJointPositionDirectly",
            index=0,
            joint_positions=list(joint_positions),
        )
        self._step()

        return self._get_obs()

    def seed(self, seed=1234):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode="human"):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        distance = self._compute_goal_distance(achieved_goal, desired_goal)
        if self.reward_type == "sparse":
            return -(distance > self.tolerance).astype(np.float32)
        else:
            return -distance

    def _get_obs(self):
        gripper_position = np.array(self.articulation_channel.data[1]["positions"][3])
        gripper_velocity = np.array(self.articulation_channel.data[1]["velocities"][3])
        gripper_width = self._get_gripper_width()

        panda_obs = np.concatenate(
            (gripper_position, gripper_velocity, [gripper_width])
        )

        grasp_position = np.array(
            self.obi_cloth_with_grasping_channel.data[0]["grasp_position"]
        )
        grasp_rotation = np.array(
            self.obi_cloth_with_grasping_channel.data[0]["grasp_rotation"]
        )
        grasp_velocity = np.array(
            self.obi_cloth_with_grasping_channel.data[0]["grasp_velocity"]
        )
        grasp_angular_vel = np.array(
            self.obi_cloth_with_grasping_channel.data[0]["grasp_angular_vel"]
        )
        achieved_goal = grasp_position.copy()

        object_obs = np.concatenate(
            (grasp_position, grasp_rotation, grasp_velocity, grasp_angular_vel)
        )

        obs = np.concatenate((panda_obs, object_obs))

        return {
            "observation": obs.copy(),
            "achieved_goal": achieved_goal.copy(),
            "desired_goal": self.goal.copy(),
        }

    def _generate_random_float(self, min: float, max: float) -> float:
        assert min < max, "Min value is {}, while max value is {}.".format(min, max)
        random_float = np.random.rand()
        random_float = random_float * (max - min) + min

        return random_float

    def _set_franka_joints(self, a: np.ndarray):
        self.articulation_channel.set_action(
            "SetJointPosition",
            index=0,
            joint_positions=list(a[0:7]),
        )
        self._step()

        a[7] = -1 * a[7] / 2
        self.articulation_channel.set_action(
            "SetJointPosition",
            index=1,
            joint_positions=[a[7], a[7]],
        )
        self._step()

    def _get_gripper_width(self):
        gripper_joint_positions = copy.deepcopy(
            self.articulation_channel.data[1]["joint_positions"]
        )
        return -1 * (gripper_joint_positions[0] + gripper_joint_positions[1])

    def _check_success(self, obs):
        achieved_goal = obs["achieved_goal"]
        desired_goal = obs["desired_goal"]
        distance = self._compute_goal_distance(achieved_goal, desired_goal)

        return (distance < self.tolerance).astype(np.float32)

    def _compute_goal_distance(self, goal_a, goal_b):
        assert goal_a.shape == goal_b.shape
        return np.linalg.norm(goal_a - goal_b, axis=-1)

    def _load_object(self):
        assert (
            self.asset_bundle_file is not None
        ), "There must be an asset bundle file to load."

        object_name = "Obi Solver For Grasp"
        position = self.np_random.uniform(self.object_range_low, self.object_range_high)
        rotation = [
            0,
            self._generate_random_float(
                -self.object_rotation_range, self.object_rotation_range
            ),
            0,
        ]

        self.asset_channel.set_action(
            "LoadObiClothWithGrasping",
            filename=self.asset_bundle_file,
            name=object_name,
            position=position,
            rotation=rotation,
        )
        self._step()
        self.has_loaded_object = True
        self.goal = np.array(
            self.obi_cloth_with_grasping_channel.data[0]["target_position"]
        )

    def _destroy_object(self):
        self.obi_cloth_with_grasping_channel.set_action("Destroy", index=0)
        self._step()
        self.has_loaded_object = False
        self.goal = None

    def _get_grasp_position(self):
        assert self.has_loaded_object, "No object loaded."
        grasp_position = np.array(
            self.obi_cloth_with_grasping_channel.data[0]["grasp_position"]
        )

        return grasp_position
